"""The `middleware_factory` method for handling the call response."""

from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from contextlib import AbstractContextManager, contextmanager
from functools import wraps
from typing import (
    Any,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from pydantic import BaseModel

from mirascope.core.base._utils._base_type import BaseType

from ..core.base._utils import fn_is_async
from ..core.base.call_response import BaseCallResponse
from ..core.base.stream import BaseStream
from ..core.base.structured_stream import BaseStructuredStream

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseStructuredStreamT = TypeVar("_BaseStructuredStreamT", bound=BaseStructuredStream)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)
ResponseModel = BaseModel | BaseType
_P = ParamSpec("_P")
SyncFunc = Callable[
    _P, _BaseCallResponseT | _BaseStreamT | _ResponseModelT | _BaseStructuredStreamT
]
AsyncFunc = Callable[
    _P,
    Awaitable[
        _BaseCallResponseT | _BaseStreamT | _ResponseModelT | _BaseStructuredStreamT
    ],
]
_T = TypeVar("_T")
_R = TypeVar("_R")


@contextmanager
def default_context_manager(
    fn: SyncFunc | AsyncFunc,
) -> Generator[None, None, None]:
    yield None


def middleware_factory(
    *,
    custom_context_manager: Callable[
        [SyncFunc | AsyncFunc], AbstractContextManager[_T]
    ] = default_context_manager,
    custom_decorator: Callable | None = None,
    handle_call_response: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_call_response_async: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_stream: Callable[[BaseStream, SyncFunc | AsyncFunc, _T | None], None]
    | None = None,
    handle_stream_async: Callable[
        [BaseStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_response_model: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_response_model_async: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_structured_stream: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_structured_stream_async: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_error: Callable[[Exception, SyncFunc | AsyncFunc, _T | None], Any]
    | None = None,
    handle_error_async: Callable[
        [Exception, SyncFunc | AsyncFunc, _T | None], Awaitable[Any]
    ]
    | None = None,
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    '''A factory method for creating middleware decorators.

    Example:

    ```python
    from mirascope.core import openai, prompt_template
    from mirascope.integrations import middleware_factory

    def with_saving():
        """Saves some data after a Mirascope call."""

        return middleware_factory(
            custom_context_manager=custom_context_manager,
            custom_decorator=custom_decorator,
            handle_call_response=handle_call_response,
            handle_call_response_async=handle_call_response_async,
            handle_stream=handle_stream,
            handle_stream_async=handle_stream_async,
            handle_response_model=handle_response_model,
            handle_response_model_async=handle_response_model_async,
            handle_structured_stream=handle_structured_stream,
            handle_structured_stream_async=handle_structured_stream_async,
            handle_error_async=handle_error_async,
            handle_error=handle_error,
        )


    @with_saving()
    @openai.call("gpt-4o-mini")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")  # `with_saving` automatically run
    print(response.content)
    ```
    '''

    @overload
    def decorator(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...

    @overload
    def decorator(fn: Callable[_P, Awaitable[_R]]) -> Callable[_P, Awaitable[_R]]: ...

    def decorator(
        fn: Callable[_P, _R | Awaitable[_R]],
    ) -> Callable[_P, _R | Awaitable[_R]]:
        if fn_is_async(fn):

            @wraps(fn)
            async def wrapper_async(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                context_manager = custom_context_manager(fn)
                context = context_manager.__enter__()
                try:
                    result = await fn(*args, **kwargs)  # pyright: ignore [reportGeneralTypeIssues]
                except Exception as e:
                    if handle_error_async:
                        try:
                            result = await handle_error_async(e, fn, context)
                        except Exception as new_e:
                            # If handle_error_async raises an exception, exit context manager and re-raise
                            context_manager.__exit__(
                                type(new_e), new_e, new_e.__traceback__
                            )
                            raise
                        else:
                            # Exception was handled, exit context manager and continue
                            context_manager.__exit__(None, None, None)
                    else:
                        # No handle_error_async provided, exit context manager with exception and re-raise
                        context_manager.__exit__(type(e), e, e.__traceback__)
                        raise

                if (
                    isinstance(result, BaseCallResponse)
                    and handle_call_response_async is not None
                ):
                    await handle_call_response_async(result, fn, context)
                    context_manager.__exit__(None, None, None)
                elif isinstance(result, BaseStream):
                    original_class = type(result)
                    original_aiter = result.__aiter__

                    def new_stream_aiter(
                        self: Any,  # noqa: ANN401
                    ) -> AsyncGenerator[tuple[Any, Any | None], Any]:  # noqa: ANN401
                        async def generator() -> AsyncGenerator[
                            tuple[Any, Any | None], Any
                        ]:
                            try:
                                async for chunk, tool in original_aiter():
                                    yield chunk, tool
                            except Exception as e:
                                if handle_error_async:
                                    try:
                                        await handle_error_async(e, fn, context)
                                    except Exception as new_e:
                                        context_manager.__exit__(
                                            type(new_e), new_e, new_e.__traceback__
                                        )
                                        raise
                                    else:
                                        context_manager.__exit__(None, None, None)
                                        return
                                else:
                                    context_manager.__exit__(
                                        type(e), e, e.__traceback__
                                    )
                                    raise
                            finally:
                                if handle_stream_async is not None:
                                    await handle_stream_async(result, fn, context)
                                context_manager.__exit__(None, None, None)

                        return generator()

                    class MiddlewareAsyncStream(original_class):
                        __aiter__ = (
                            custom_decorator(fn)(new_stream_aiter)
                            if custom_decorator
                            else new_stream_aiter
                        )

                    result.__class__ = MiddlewareAsyncStream
                elif (
                    isinstance(result, ResponseModel)
                    and handle_response_model_async is not None
                ):
                    await handle_response_model_async(result, fn, context)
                    context_manager.__exit__(None, None, None)
                elif isinstance(result, BaseStructuredStream):
                    original_class = type(result)
                    original_aiter = result.__aiter__

                    def new_aiter(
                        self: Any,  # noqa: ANN401
                    ) -> AsyncGenerator[tuple[Any, Any | None], Any]:  # noqa: ANN401
                        async def generator() -> AsyncGenerator[
                            tuple[Any, Any | None], Any
                        ]:
                            try:
                                async for chunk in original_aiter():
                                    yield chunk
                            except Exception as e:
                                if handle_error_async:
                                    try:
                                        await handle_error_async(e, fn, context)
                                    except Exception as new_e:
                                        context_manager.__exit__(
                                            type(new_e), new_e, new_e.__traceback__
                                        )
                                        raise
                                    else:
                                        context_manager.__exit__(None, None, None)
                                        return
                                else:
                                    context_manager.__exit__(
                                        type(e), e, e.__traceback__
                                    )
                                    raise
                            finally:
                                if handle_structured_stream_async is not None:
                                    await handle_structured_stream_async(
                                        result, fn, context
                                    )
                                context_manager.__exit__(None, None, None)

                        return generator()

                    class MiddlewareAsyncStructuredStream(original_class):
                        __aiter__ = (
                            custom_decorator(fn)(new_aiter)
                            if custom_decorator
                            else new_aiter
                        )

                    result.__class__ = MiddlewareAsyncStructuredStream
                else:
                    context_manager.__exit__(None, None, None)
                return cast(_R, result)

            return (
                custom_decorator(fn)(wrapper_async)
                if custom_decorator
                else wrapper_async
            )
        else:

            @wraps(fn)
            def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                context_manager = custom_context_manager(fn)
                context = context_manager.__enter__()
                try:
                    result = fn(*args, **kwargs)
                except Exception as e:
                    if handle_error:
                        try:
                            result = handle_error(e, fn, context)
                        except Exception as new_e:
                            # If handle_error raises an exception, exit context manager and re-raise
                            context_manager.__exit__(
                                type(new_e), new_e, new_e.__traceback__
                            )
                            raise
                        else:
                            # Exception was handled, exit context manager and continue
                            context_manager.__exit__(None, None, None)
                    else:
                        # No handle_error provided, exit context manager with exception and re-raise
                        context_manager.__exit__(type(e), e, e.__traceback__)
                        raise

                if (
                    isinstance(result, BaseCallResponse)
                    and handle_call_response is not None
                ):
                    handle_call_response(result, fn, context)
                    context_manager.__exit__(None, None, None)
                elif isinstance(result, BaseStream):
                    original_class = type(result)
                    original_iter = result.__iter__

                    def new_stream_iter(
                        self: Any,  # noqa: ANN401
                    ) -> Generator[tuple[Any, Any | None], None, None]:  # noqa: ANN401
                        # Create the context manager when user iterates over the stream
                        try:
                            yield from original_iter()
                        except Exception as e:
                            if handle_error:
                                try:
                                    handle_error(e, fn, context)
                                except Exception as new_e:
                                    context_manager.__exit__(
                                        type(new_e), new_e, new_e.__traceback__
                                    )
                                    raise
                                else:
                                    context_manager.__exit__(None, None, None)
                                    return
                            else:
                                context_manager.__exit__(type(e), e, e.__traceback__)
                                raise
                        finally:
                            if handle_stream is not None:
                                handle_stream(result, fn, context)
                            context_manager.__exit__(None, None, None)

                    class MiddlewareStream(original_class):  # pyright: ignore [reportGeneralTypeIssues]
                        __iter__ = (
                            custom_decorator(fn)(new_stream_iter)
                            if custom_decorator
                            else new_stream_iter
                        )

                    result.__class__ = MiddlewareStream
                elif (
                    isinstance(result, ResponseModel)
                    and handle_response_model is not None
                ):
                    handle_response_model(result, fn, context)
                    context_manager.__exit__(None, None, None)
                elif isinstance(result, BaseStructuredStream):
                    original_class = type(result)
                    original_iter = result.__iter__

                    def new_iter(
                        self: Any,  # noqa: ANN401
                    ) -> Generator[
                        Generator[tuple[Any, Any | None], None, None]
                        | Generator[Any, None, None],
                        Any,
                        None,
                    ]:  # noqa: ANN401
                        # Create the context manager when user iterates over the stream
                        try:
                            yield from original_iter()
                        except Exception as e:
                            if handle_error:
                                try:
                                    handle_error(e, fn, context)
                                except Exception as new_e:
                                    context_manager.__exit__(
                                        type(new_e), new_e, new_e.__traceback__
                                    )
                                    raise
                                else:
                                    context_manager.__exit__(None, None, None)
                                    return
                            else:
                                context_manager.__exit__(type(e), e, e.__traceback__)
                                raise
                        finally:
                            if handle_structured_stream is not None:
                                handle_structured_stream(result, fn, context)
                            context_manager.__exit__(None, None, None)

                    class MiddlewareStructuredStream(original_class):  # pyright: ignore [reportGeneralTypeIssues]
                        __iter__ = (
                            custom_decorator(fn)(new_iter)
                            if custom_decorator
                            else new_iter
                        )

                    result.__class__ = MiddlewareStructuredStream
                else:
                    context_manager.__exit__(None, None, None)
                return cast(_R, result)

            return custom_decorator(fn)(wrapper) if custom_decorator else wrapper

    return decorator



"""Tests the `groq._utils.handle_stream` module."""

import pytest
from groq.types.chat import ChatCompletionChunk
from groq.types.chat.chat_completion_chunk import (
    Choice,
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)

from mirascope.core.groq._utils._handle_stream import (
    handle_stream,
    handle_stream_async,
)
from mirascope.core.groq.tool import GroqTool


class FormatBook(GroqTool):
    """Returns the title and author nicely formatted."""

    title: str
    author: str

    def call(self) -> None:
        """Dummy call."""


@pytest.fixture()
def mock_chunks() -> list[ChatCompletionChunk]:
    """Returns a list of mock `ChatCompletionChunk` instances."""

    new_tool_call = ChoiceDeltaToolCall(
        index=0,
        id="id",
        function=ChoiceDeltaToolCallFunction(
            arguments=None,
            name="FormatBook",
        ),
        type="function",
    )
    tool_call = ChoiceDeltaToolCall(
        index=0,
        id=None,
        function=ChoiceDeltaToolCallFunction(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name=None,
        ),
        type="function",
    )
    return [
        ChatCompletionChunk(
            id="id",
            choices=[
                Choice(delta=ChoiceDelta(content="content", tool_calls=None), index=0)
            ],
            created=0,
            model="llama3-groq-70b-8192-tool-use-preview",
            object="chat.completion.chunk",
            x_groq=None,
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        tool_calls=[new_tool_call],
                    ),
                    index=0,
                )
            ],
            created=0,
            model="llama3-groq-70b-8192-tool-use-preview",
            object="chat.completion.chunk",
            x_groq=None,
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    index=0,
                )
            ],
            created=0,
            model="llama3-groq-70b-8192-tool-use-preview",
            object="chat.completion.chunk",
            x_groq=None,
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        tool_calls=[new_tool_call],
                    ),
                    index=0,
                )
            ],
            created=0,
            model="llama3-groq-70b-8192-tool-use-preview",
            object="chat.completion.chunk",
            x_groq=None,
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    index=0,
                )
            ],
            created=0,
            model="llama3-groq-70b-8192-tool-use-preview",
            object="chat.completion.chunk",
            x_groq=None,
        ),
        ChatCompletionChunk(
            id="id",
            choices=[
                Choice(
                    delta=ChoiceDelta(content=None, tool_calls=None),
                    finish_reason="tool_calls",
                    index=0,
                )
            ],
            created=0,
            model="llama3-groq-70b-8192-tool-use-preview",
            object="chat.completion.chunk",
            x_groq=None,
        ),
    ]


def test_handle_stream(mock_chunks: list[ChatCompletionChunk]) -> None:
    """Tests the `handle_stream` function."""

    result = list(handle_stream((c for c in mock_chunks), tool_types=[FormatBook]))
    # Check we get three tuples back.
    # (chunk, None), (chunk, FormatBook), (chunk, FormatBook)
    assert len(result) == 3
    assert result[0][1] is None
    assert (
        (tool := result[1][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )
    assert (
        (tool := result[2][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )


@pytest.mark.asyncio
async def test_handle_stream_async(mock_chunks: list[ChatCompletionChunk]) -> None:
    """Tests the `handle_stream_async` function."""

    async def generator():
        for chunk in mock_chunks:
            yield chunk

    result = []
    async for t in handle_stream_async(generator(), tool_types=[FormatBook]):
        result.append(t)
    # Check we get three tuples back.
    # (chunk, None), (chunk, FormatBook), (chunk, FormatBook)
    assert len(result) == 3
    assert result[0][1] is None
    assert (
        (tool := result[1][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )
    assert (
        (tool := result[2][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )



"""This module contains the setup_call function for Bedrock tools."""

from __future__ import annotations

import os
from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine, Generator
from functools import wraps
from typing import Any, ParamSpec, cast, overload

import aiobotocore.client
from aiobotocore.session import AioSession, get_session
from boto3.session import Session
from mypy_boto3_bedrock_runtime import BedrockRuntimeClient
from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef,
    ConverseStreamResponseTypeDef,
    ToolChoiceTypeDef,
    ToolConfigurationTypeDef,
)
from pydantic import BaseModel
from types_aiobotocore_bedrock_runtime import (
    BedrockRuntimeClient as AsyncBedrockRuntimeClient,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseRequestRequestTypeDef as AsyncConverseRequestRequestTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef as AsyncConverseResponseTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseStreamRequestRequestTypeDef as AsyncConverseStreamRequestRequestTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseStreamResponseTypeDef as AsyncConverseStreamResponseTypeDef,
)
from typing_extensions import Unpack

from ... import BaseMessageParam
from ...base import BaseTool, _utils
from ...base._utils import (
    AsyncCreateFn,
    CreateFn,
    fn_is_async,
    get_async_create_fn,
    get_create_fn,
)
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import BedrockCallKwargs
from .._types import (
    AsyncStreamOutputChunk,
    InternalBedrockMessageParam,
    StreamOutputChunk,
)
from ..call_params import BedrockCallParams
from ..dynamic_config import AsyncBedrockDynamicConfig, BedrockDynamicConfig
from ..tool import BedrockTool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params

_P = ParamSpec("_P")


def _extract_sync_stream_fn(
    fn: Callable[_P, ConverseStreamResponseTypeDef], model: str
) -> Callable[_P, Generator[StreamOutputChunk, None, None]]:
    @wraps(fn)
    def _inner(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> Generator[StreamOutputChunk, None, None]:
        response = fn(*args, **kwargs)
        for chunk in response["stream"]:
            yield StreamOutputChunk(
                responseMetadata=response["ResponseMetadata"], model=model, **chunk
            )

    return _inner


def _extract_async_stream_fn(
    fn: Callable[_P, Coroutine[Any, Any, AsyncConverseStreamResponseTypeDef]],
    model: str,
) -> Callable[_P, AsyncGenerator[AsyncStreamOutputChunk, None]]:
    @wraps(fn)
    async def _inner(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> AsyncGenerator[AsyncStreamOutputChunk, None]:
        response = await fn(*args, **kwargs)
        async for chunk in response["stream"]:
            yield AsyncStreamOutputChunk(
                responseMetadata=response["ResponseMetadata"], model=model, **chunk
            )

    return _inner


class _AsyncBedrockRuntimeWrappedClient:
    def __init__(self, session: AioSession, model: str) -> None:
        self.session: AioSession = session
        self.model: str = model

    async def converse(
        self, **kwargs: Unpack[AsyncConverseRequestRequestTypeDef]
    ) -> AsyncConverseResponseTypeDef:
        async with self.session.create_client("bedrock-runtime") as client:
            return await client.converse(**kwargs)

    async def converse_stream(
        self, **kwargs: Unpack[AsyncConverseStreamRequestRequestTypeDef]
    ) -> AsyncGenerator[AsyncStreamOutputChunk, None]:
        async with self.session.create_client("bedrock-runtime") as client:
            response = await client.converse_stream(**kwargs)
            async for chunk in response["stream"]:
                yield AsyncStreamOutputChunk(
                    responseMetadata=response["ResponseMetadata"],
                    model=self.model,
                    **chunk,
                )


@overload
def setup_call(
    *,
    model: str,
    client: AsyncBedrockRuntimeClient | None,
    fn: Callable[..., Awaitable[AsyncBedrockDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncBedrockDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: BedrockCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[AsyncConverseResponseTypeDef, AsyncStreamOutputChunk],
    str | None,
    list[InternalBedrockMessageParam],
    list[type[BedrockTool]] | None,
    BedrockCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: BedrockRuntimeClient | None,
    fn: Callable[..., BedrockDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: BedrockDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: BedrockCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ConverseResponseTypeDef, StreamOutputChunk],
    str | None,
    list[InternalBedrockMessageParam],
    list[type[BedrockTool]] | None,
    BedrockCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: BedrockRuntimeClient | AsyncBedrockRuntimeClient | None,
    fn: Callable[..., BedrockDynamicConfig]
    | Callable[..., Awaitable[AsyncBedrockDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: BedrockDynamicConfig | AsyncBedrockDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: BedrockCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[AsyncConverseResponseTypeDef, AsyncStreamOutputChunk]
    | CreateFn[ConverseResponseTypeDef, StreamOutputChunk],
    str | None,
    list[InternalBedrockMessageParam],
    list[type[BedrockTool]] | None,
    BedrockCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        BedrockTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(BedrockCallKwargs, base_call_kwargs)
    messages = cast(list[InternalBedrockMessageParam | BaseMessageParam], messages)
    messages = convert_message_params(messages)
    if messages[0]["role"] == "system":
        call_kwargs["system"] = [
            {"text": text}
            for c in messages.pop(0)["content"]
            if (text := c.get("text"))
        ]

    if call_kwargs_tools := call_kwargs.pop("tools", None):
        call_kwargs["toolConfig"] = cast(
            ToolConfigurationTypeDef, {"tools": call_kwargs_tools}
        )
    if json_mode:
        json_mode_content = _utils.json_mode_content(response_model)
        if "text" in messages[-1]["content"][-1]:
            messages[-1]["content"][-1]["text"] += json_mode_content
        else:
            messages[-1]["content"] = [
                *messages[-1]["content"],
                {"text": json_mode_content},
            ]
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        if "toolConfig" in call_kwargs:
            call_kwargs["toolConfig"]["toolChoice"] = cast(
                ToolChoiceTypeDef, {"tool": {"name": tool_types[0]._name()}}
            )

    call_kwargs |= cast(BedrockCallKwargs, {"modelId": model, "messages": messages})

    env_vars = {}
    if access_key_id := os.getenv("AWS_ACCESS_KEY_ID"):
        env_vars["aws_access_key_id"] = access_key_id
    if secret_access_key := os.getenv("AWS_SECRET_ACCESS_KEY"):
        env_vars["aws_secret_access_key"] = secret_access_key
    if session_token := os.getenv("AWS_SESSION_TOKEN"):
        env_vars["aws_session_token"] = session_token
    if region_name := os.getenv("AWS_REGION_NAME"):
        env_vars["region_name"] = region_name
    if profile_name := os.getenv("AWS_PROFILE"):
        env_vars["profile_name"] = profile_name
    if client is None:
        if fn_is_async(fn):
            session = get_session(env_vars=env_vars)
            _client = _AsyncBedrockRuntimeWrappedClient(session, model)
        else:
            session = Session(**env_vars)
            _client = session.client("bedrock-runtime")
    else:
        _client = client
    if isinstance(_client, aiobotocore.client.AioBaseClient):
        create = get_async_create_fn(
            _client.converse, _extract_async_stream_fn(_client.converse_stream, model)
        )
    elif isinstance(_client, _AsyncBedrockRuntimeWrappedClient):
        create = get_async_create_fn(_client.converse, _client.converse_stream)

    else:
        create = get_create_fn(
            _client.converse, _extract_sync_stream_fn(_client.converse_stream, model)
        )
    return create, prompt_template, messages, tool_types, call_kwargs



"""This module defines the function return type for functions as LLM calls."""

from mypy_boto3_bedrock_runtime import BedrockRuntimeClient
from types_aiobotocore_bedrock_runtime import (
    BedrockRuntimeClient as AsyncBedrockRuntimeClient,
)

from ..base import BaseDynamicConfig, BaseMessageParam
from ._types import InternalBedrockMessageParam
from .call_params import BedrockCallParams

AsyncBedrockDynamicConfig = BaseDynamicConfig[
    InternalBedrockMessageParam | BaseMessageParam,
    BedrockCallParams,
    AsyncBedrockRuntimeClient,
]
BedrockDynamicConfig = BaseDynamicConfig[
    InternalBedrockMessageParam | BaseMessageParam,
    BedrockCallParams,
    BedrockRuntimeClient,
]

"""The function return type for functions wrapped with the `bedrock_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.bedrock import BedrockDynamicConfig, bedrock_call


@bedrock_call("anthropic.claude-3-haiku-20240307-v1:0")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> BedrockDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""The `anthropic_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import AnthropicCallParams
from .call_response import AnthropicCallResponse
from .call_response_chunk import AnthropicCallResponseChunk
from .stream import AnthropicStream
from .tool import AnthropicTool

anthropic_call = call_factory(
    TCallResponse=AnthropicCallResponse,
    TCallResponseChunk=AnthropicCallResponseChunk,
    TStream=AnthropicStream,
    TToolType=AnthropicTool,
    default_call_params=AnthropicCallParams(max_tokens=1024),
    setup_call=setup_call,
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
"""A decorator for calling the Anthropic API with a typed function.

usage docs: learn/calls.md

This decorator is used to wrap a typed function that calls the Anthropic API. It
parses the prompt template of the wrapped function as the messages array and templates
the input arguments for the function into each message's template.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.anthropic import anthropic_call


@anthropic_call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"

response = recommend_book("fantasy")
print(response.content)
```

Args:
    model (str): The Anthropic model to use in the API call.
    stream (bool): Whether to stream the response from the API call.
    tools (list[BaseTool | Callable]): The tools to use in the Anthropic API call.
    response_model (BaseModel | BaseType): The response model into which the response
        should be structured.
    output_parser (Callable[[AnthropicCallResponse | ResponseModelT], Any]): A function
        for parsing the call response whose value will be returned in place of the
        original call response.
    json_mode (bool): Whether to use JSON Mode.
    client (object): An optional custom client to use in place of the default client.
    call_params (AnthropicCallParams): The `AnthropicCallParams` call parameters to use
        in the API call.

Returns:
    decorator (Callable): The decorator for turning a typed function into an Anthropic
        API call.
"""



from typing import ClassVar, Literal

import requests
from pydantic import Field

from mirascope.tools.base import ConfigurableTool, _ConfigurableToolConfig


class RequestsConfig(_ConfigurableToolConfig):
    """Configuration for HTTP requests"""

    timeout: int = Field(default=5, description="Request timeout in seconds")


class Requests(ConfigurableTool):
    """Tool for making HTTP requests with built-in requests library."""

    __configurable_tool_config__ = RequestsConfig()

    __prompt_usage_description__: ClassVar[str] = """
    - `Requests`: Makes HTTP requests to web URLs using requests library
        - Supports common HTTP methods (GET, POST, PUT, DELETE)
        - Accepts custom headers and request data
        - Returns response text content
        - Automatically raises for HTTP errors (4xx, 5xx)
        - Returns error message if request fails
        - Configurable timeout for all requests
    """
    url: str = Field(..., description="URL to request")
    method: Literal["GET", "POST", "PUT", "DELETE"] = Field(
        "GET", description="HTTP method"
    )
    data: dict | None = Field(None, description="Data to send with POST/PUT requests")
    headers: dict | None = Field(None, description="Request headers")

    def call(self) -> str:
        """Make an HTTP request to the given URL.

        Returns:
            str: Response text content if successful, error message if request fails
        """
        try:
            response = requests.request(
                method=self.method,
                url=self.url,
                json=self.data,
                headers=self.headers,
                timeout=self._get_config().timeout,
            )
            response.raise_for_status()
            return response.text

        except Exception as e:
            return f"{type(e)}: Failed to extract content from URL {self.url}"



"""Tests the `vertex.stream` module."""

import pytest
from vertexai.generative_models import (
    Candidate,
    Content,
    FinishReason,
    Part,
)
from vertexai.generative_models import (
    GenerationResponse as GenerateContentResponse,
)

from mirascope.core.base.types import CostMetadata
from mirascope.core.vertex.call_response import VertexCallResponse
from mirascope.core.vertex.call_response_chunk import VertexCallResponseChunk
from mirascope.core.vertex.stream import VertexStream


def test_vertex_stream() -> None:
    """Tests the `VertexStream` class."""
    assert VertexStream._provider == "vertex"

    chunks = [
        GenerateContentResponse.from_dict(
            {
                "candidates": [
                    Candidate.from_dict(
                        {
                            "finish_reason": 1,
                            "content": Content(
                                parts=[Part.from_text("The author is ")],
                                role="model",
                            ).to_dict(),
                        },
                    ).to_dict()
                ]
            }
        ),
        GenerateContentResponse.from_dict(
            {
                "candidates": [
                    Candidate.from_dict(
                        {
                            "finish_reason": 1,
                            "content": Content(
                                parts=[Part.from_text("Patrick Rothfuss")], role="model"
                            ).to_dict(),
                        }
                    ).to_dict()
                ]
            }
        ),
    ]
    stream = VertexStream(
        stream=((VertexCallResponseChunk(chunk=chunk), None) for chunk in chunks),
        metadata={},
        tool_types=None,
        call_response_type=VertexCallResponse,
        model="gemini-1.5-flash",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[Content(role="user", parts=[Part.from_text("Who is the author?")])],
        call_params={},
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()

    assert stream.cost is None
    for _ in stream:
        pass
    assert stream.cost is None
    assert stream.message_param.role == "model"
    assert len(stream.message_param.parts) == 1
    assert stream.message_param.parts[0].text == "The author is Patrick Rothfuss"
    assert stream.cost_metadata == CostMetadata()


def test_construct_call_response() -> None:
    chunks = [
        GenerateContentResponse.from_dict(
            {
                "candidates": [
                    Candidate.from_dict(
                        {
                            "content": Content(
                                parts=[Part.from_text("The author is ")], role="model"
                            ).to_dict(),
                            "finish_reason": FinishReason.STOP,
                        }
                    ).to_dict()
                ]
            }
        ),
        GenerateContentResponse.from_dict(
            {
                "candidates": [
                    Candidate.from_dict(
                        {
                            "content": Content(
                                parts=[Part.from_text("Patrick Rothfuss")], role="model"
                            ).to_dict(),
                            "finish_reason": FinishReason.STOP,
                        }
                    ).to_dict()
                ]
            }
        ),
    ]
    stream = VertexStream(
        stream=((VertexCallResponseChunk(chunk=chunk), None) for chunk in chunks),
        metadata={},
        tool_types=None,
        call_response_type=VertexCallResponse,
        model="gemini-1.5-flash",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[Content(role="user", parts=[Part.from_text("Who is the author?")])],
        call_params={},
        call_kwargs={},
    )

    for _ in stream:
        pass

    expected_response = GenerateContentResponse.from_dict(
        {
            "candidates": [
                Candidate.from_dict(
                    {
                        "content": Content(
                            parts=[Part.from_text("The author is Patrick Rothfuss")],
                            role="model",
                        ).to_dict(),
                        "finish_reason": FinishReason.STOP,
                    }
                ).to_dict()
            ]
        }
    )
    call_response = VertexCallResponse(
        metadata={},
        response=expected_response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    constructed_call_response = stream.construct_call_response()
    assert (
        constructed_call_response.response.to_dict() == call_response.response.to_dict()
    )



"""Tests the `xai.stream` module."""

from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice as ChunkChoice
from openai.types.chat.chat_completion_chunk import ChoiceDelta
from openai.types.completion_usage import CompletionUsage

from mirascope.core.xai.call_response import XAICallResponse
from mirascope.core.xai.call_response_chunk import XAICallResponseChunk
from mirascope.core.xai.stream import XAIStream


def test_xai_stream_cost() -> None:
    """Tests the `XAIStream` cost property."""
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    chunks = [
        ChatCompletionChunk(
            id="id",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(content="content", tool_calls=None), index=0
                )
            ],
            created=0,
            model="grok-3",
            object="chat.completion.chunk",
            usage=usage,
        ),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = XAICallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = XAIStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=XAICallResponse,
        model="",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params={},
        call_kwargs={},
    )

    for _ in stream:
        pass

    assert stream.cost == 1.4e-5
    assert stream.model == "grok-3"



"""Tests the `google.tool` module."""

from enum import Enum
from typing import Literal

from google.genai.types import (  # type: ignore
    FunctionCall,
    FunctionDeclaration,
    Schema,
    Tool,
    Type,
)
from pydantic import Field

from mirascope.core.base.tool import BaseTool
from mirascope.core.google.tool import GoogleTool


class FormatBook(GoogleTool):
    """Returns the title and author nicely formatted."""

    title: str = Field(..., examples=["The Name of the Wind"])
    author: str
    genre: Literal["fantasy", "scifi"]

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "title": "The Way of Kings",
                    "author": "Brandon Sanderson",
                    "genre": "fantasy",
                },
                {"title": "Dune", "author": "Frank Herbert", "genre": "scifi"},
            ]
        }
    }

    def call(self) -> str:
        return f"{self.title} by {self.author}"


def test_google_tool() -> None:
    """Tests the `GoogleTool` class."""
    tool_call = FunctionCall(
        name="FormatBook",
        args={
            "title": "The Name of the Wind",
            "author": "Patrick Rothfuss",
            "genre": "fantasy",
        },
    )

    tool = FormatBook.from_tool_call(tool_call)
    assert isinstance(tool, BaseTool)
    assert isinstance(tool, GoogleTool)
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"

    assert FormatBook.tool_schema() == Tool(
        function_declarations=[
            FunctionDeclaration(
                response=None,
                description="Returns the title and author nicely formatted.",
                name="FormatBook",
                parameters=Schema(
                    type=Type.OBJECT,
                    properties={
                        "title": Schema(
                            type=Type.STRING, example=["The Name of the Wind"]
                        ),
                        "author": Schema(type=Type.STRING),
                        "genre": Schema(
                            type=Type.STRING,
                            enum=["fantasy", "scifi"],
                            format="enum",
                        ),
                    },
                    required=["title", "author", "genre"],
                    example=[
                        {
                            "title": "The Way of Kings",
                            "author": "Brandon Sanderson",
                            "genre": "fantasy",
                        },
                        {"title": "Dune", "author": "Frank Herbert", "genre": "scifi"},
                    ],
                ),
            )
        ],
    )


def test_google_tool_no_nesting() -> None:
    """Tests the `GoogleTool` class with nested structures."""

    class Def(Enum):
        A = "a"

    class Nested(GoogleTool):
        """Nested tool."""

        nested: Def

    assert Nested.tool_schema() == Tool(
        function_declarations=[
            FunctionDeclaration(
                response=None,
                description="Nested tool.",
                name="Nested",
                parameters=Schema(
                    type=Type.OBJECT,
                    properties={
                        "nested": Schema(type=Type.STRING, enum=["a"], format="enum")
                    },
                    required=["nested"],
                ),
            )
        ]
    )



"""Tests the `_utils.get_image_type` module."""

import pytest

from mirascope.core.base._utils._get_image_type import get_image_type

JPEG_DATA = b"\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x02\x01\x00e\x00e\x00\x00\xff\xe1\x10>Exif\x00\x00"

PNG_DATA = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x03 \x00\x00\x02X\x08\x06\x00\x00\x00\x9a"

GIF_DATA = b"GIF89a\x90\x01\x90\x01\xf7\x00\x00\x00\x00\x00\x00\x009\x00\x00A\x00\x001\x00\x00\x08\x00\x00"

WEBP_DATA = b"RIFFhv\x00\x00WEBPVP8 \\v\x00\x00\xd2\xbe\x01\x9d\x01*&\x02p\x01"

HEIC_DATA = b"\x00\x00\x00\x18ftypheic\x00\x00\x00\x00mif1heic\x00\x00\x01\xb7me"

HEIF_DATA = b"\x00\x00\x00\x18ftyphevx\x00\x00\x00\x00mif1heic\x00\x00\x01\xb7me"


def test_get_image_type() -> None:
    """Test the get_image_type function."""
    assert get_image_type(JPEG_DATA) == "jpeg"
    assert get_image_type(PNG_DATA) == "png"
    assert get_image_type(GIF_DATA) == "gif"
    assert get_image_type(WEBP_DATA) == "webp"
    assert get_image_type(HEIC_DATA) == "heic"
    assert get_image_type(HEIF_DATA) == "heif"
    with pytest.raises(ValueError, match="Unsupported image type"):
        get_image_type(b"invalid")



"""Utility for converting `BaseMessageParam` to `ChatCompletionMessageParam`."""

import base64
import json

from openai.types.chat import ChatCompletionMessageParam

from ...base import BaseMessageParam
from ...base._utils import get_audio_type
from ...base._utils._parse_content_template import _load_media


def convert_message_params(
    message_params: list[BaseMessageParam | ChatCompletionMessageParam],
) -> list[ChatCompletionMessageParam]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance((content := message_param.content), str):
            converted_message_params.append(message_param.model_dump())
        else:
            converted_content = []

            tool_calls = []
            for part in content:
                if part.type == "text":
                    converted_content.append(part.model_dump())
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/gif",
                        "image/webp",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. OpenAI"
                            " currently only supports JPEG, PNG, GIF, and WebP images."
                        )
                    data = base64.b64encode(part.image).decode("utf-8")
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:{part.media_type};base64,{data}",
                                "detail": part.detail if part.detail else "auto",
                            },
                        }
                    )
                elif part.type == "image_url":
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": part.url,
                                "detail": part.detail if part.detail else "auto",
                            },
                        }
                    )
                elif part.type == "audio":
                    if part.media_type not in [
                        "audio/wav",
                        "audio/mp3",
                    ]:
                        raise ValueError(
                            f"Unsupported audio media type: {part.media_type}. "
                            "OpenAI currently only supports WAV and MP3 audio file types."
                        )
                    data = (
                        part.audio
                        if isinstance(part.audio, str)
                        else base64.b64encode(part.audio).decode("utf-8")
                    )
                    converted_content.append(
                        {
                            "input_audio": {
                                "format": part.media_type.split("/")[-1],
                                "data": data,
                            },
                            "type": "input_audio",
                        }
                    )
                elif part.type == "audio_url":
                    audio = _load_media(part.url)
                    audio_type = get_audio_type(audio)
                    if audio_type not in [
                        "wav",
                        "mp3",
                    ]:
                        raise ValueError(
                            f"Unsupported audio media type: audio/{audio_type}. "
                            "OpenAI currently only supports WAV and MP3 audio file types."
                        )
                    converted_content.append(
                        {
                            "input_audio": {
                                "format": audio_type,
                                "data": base64.b64encode(audio).decode("utf-8"),
                            },
                            "type": "input_audio",
                        }
                    )
                elif part.type == "tool_call":
                    tool_calls.append(
                        {
                            "function": {
                                "name": part.name,
                                "arguments": json.dumps(part.args),
                            },
                            "type": "function",
                            "id": part.id,
                        }
                    )
                elif part.type == "tool_result":
                    if converted_content:
                        converted_message_params.append(
                            {"role": message_param.role, "content": converted_content}
                        )
                        converted_content = []

                    converted_message_params.append(
                        {
                            "role": "tool",
                            "tool_call_id": part.id,
                            "content": part.content,
                        }
                    )
                else:
                    raise ValueError(
                        "OpenAI currently only supports text, image and audio parts. "
                        f"Part provided: {part.type}"
                    )
            if tool_calls:
                converted_message_param = {
                    "role": "assistant",
                    "tool_calls": tool_calls,
                }
                converted_message_params.append(converted_message_param)
                if converted_content:
                    converted_message_param["content"] = converted_content
            elif converted_content:
                converted_message_params.append(
                    {
                        "role": message_param.role,
                        "content": converted_content,
                    }
                )
    return converted_message_params



from unittest.mock import MagicMock

from cohere.types import ChatMessage

from mirascope.core import BaseMessageParam
from mirascope.core.base import TextPart, ToolCallPart
from mirascope.core.cohere._utils._message_param_converter import (
    CohereMessageParamConverter,
)


def test_no_tool_calls_with_message():
    """
    If no tool_calls are present and there's a message, return role="assistant", content=string.
    """
    message_param = MagicMock(spec=ChatMessage)
    message_param.tool_calls = None
    message_param.message = "Hello assistant"

    results = CohereMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert result.content == "Hello assistant"


def test_tool_calls_with_message():
    """
    If tool_calls are present and there's a message, we expect role="tool" with [TextPart, ToolCallPart].
    """
    tool_call = MagicMock()
    tool_call.name = "test_tool"
    tool_call.parameters = {"arg": "value"}

    message_param = MagicMock(spec=ChatMessage)
    message_param.tool_calls = [tool_call]
    message_param.message = "Message before tool"

    results = CohereMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert len(result.content) == 2
    assert isinstance(result.content[0], TextPart)
    assert result.content[0].text == "Message before tool"
    part = result.content[1]
    assert isinstance(part, ToolCallPart)
    assert part.name == "test_tool"
    assert part.args == {"arg": "value"}


def test_tool_calls_no_message():
    """
    If tool_calls are present but no message, we get multiple ToolCallParts.
    """
    tool_call1 = MagicMock()
    tool_call1.name = "tool_one"
    tool_call1.parameters = {"param1": "val1"}

    tool_call2 = MagicMock()
    tool_call2.name = "tool_two"
    tool_call2.parameters = {"param2": "val2"}

    message_param = MagicMock(spec=ChatMessage)
    message_param.tool_calls = [tool_call1, tool_call2]
    message_param.message = None

    results = CohereMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert len(result.content) == 2
    assert isinstance(result.content[0], ToolCallPart)
    assert result.content[0].name == "tool_one"
    assert result.content[0].args == {"param1": "val1"}
    assert isinstance(result.content[1], ToolCallPart)
    assert result.content[1].name == "tool_two"
    assert result.content[1].args == {"param2": "val2"}


def test_base_message_param_to_provider():
    """
    Test that `to_provider` method returns a list of ChatMessage objects.
    """
    message_param = BaseMessageParam(role="assistant", content="Hello world")
    results = CohereMessageParamConverter.to_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, ChatMessage)
    assert result.message == "Hello world"
    assert result.tool_calls is None



"""Handles streaming content and tools from the Vertex API."""

from collections.abc import AsyncGenerator, Generator

from vertexai.generative_models import GenerationResponse

from ..call_response_chunk import VertexCallResponseChunk
from ..tool import VertexTool


def handle_stream(
    stream: Generator[GenerationResponse, None, None],
    tool_types: list[type[VertexTool]] | None = None,
) -> Generator[tuple[VertexCallResponseChunk, None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed.

    Note: vertex does not currently support streaming tools.
    """
    for chunk in stream:
        yield VertexCallResponseChunk(chunk=chunk), None


async def handle_stream_async(
    stream: AsyncGenerator[GenerationResponse, None],
    tool_types: list[type[VertexTool]] | None = None,
) -> AsyncGenerator[tuple[VertexCallResponseChunk, None], None]:
    """
    Async iterator over the stream and constructs tools as they are streamed.

    Note: vertex does not currently support streaming tools.
    """
    async for chunk in stream:
        yield VertexCallResponseChunk(chunk=chunk), None



"""Tests the `call_response_chunk` module."""

from unittest.mock import patch

from mirascope.core.base.call_response_chunk import BaseCallResponseChunk


def test_base_call_response_chunk() -> None:
    class MyCallResponseChunk(BaseCallResponseChunk):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponseChunk, __abstractmethods__=set()).start()
    call_response_chunk = MyCallResponseChunk(chunk="")  # type: ignore
    assert str(call_response_chunk) == "content"



"""The `MCPServer` Class and context managers."""

import contextlib
from collections.abc import AsyncIterator, Awaitable, Callable
from datetime import timedelta
from typing import Any

from mcp import ClientSession
from mcp.client.session import ListRootsFnT, SamplingFnT
from mcp.client.sse import sse_client as mcp_sse_client
from mcp.client.stdio import StdioServerParameters
from mcp.client.stdio import stdio_client as mcp_stdio_client
from mcp.types import (
    BlobResourceContents,
    Resource,
)
from pydantic import AnyUrl

from mirascope.core import BaseMessageParam, BaseTool, TextPart
from mirascope.mcp import _utils


class MCPClient(ClientSession):
    """The SSE client session that connects to the MCP server.

    All of the results from the server are converted into Mirascope-friendly types.
    """

    _session: ClientSession

    def __init__(self, session: ClientSession) -> None:
        """Initializes an instance of `MCPClient`.

        Args:
            session: The original MCP `ClientSession`.
        """
        self._session = session

    async def list_resources(self) -> list[Resource]:  # pyright: ignore [reportIncompatibleMethodOverride]
        """List all resources available on the MCP server.

        Returns:
            A list of Resource objects
        """
        result = await self._session.list_resources()
        return result.resources

    async def read_resource(  # pyright: ignore [reportIncompatibleMethodOverride]
        self, uri: str | AnyUrl
    ) -> list[TextPart | BlobResourceContents]:
        """Read a resource from the MCP server.

        Args:
            uri: URI of the resource to read

        Returns:
            Contents of the resource, either as string or BlobResourceContents
        """
        result = await self._session.read_resource(
            uri if isinstance(uri, AnyUrl) else AnyUrl(uri)
        )
        parsed_results: list[TextPart | BlobResourceContents] = []
        for content in result.contents:
            if isinstance(content, BlobResourceContents):
                parsed_results.append(content)
            else:
                parsed_results.append(TextPart(type="text", text=content.text))
        return parsed_results

    async def list_prompts(self) -> list[Any]:  # pyright: ignore [reportIncompatibleMethodOverride]
        """List all prompts available on the MCP server.

        Returns:
            A list of Prompt objects
        """
        result = await self._session.list_prompts()
        return result.prompts

    async def get_prompt_template(
        self, name: str
    ) -> Callable[..., Awaitable[list[BaseMessageParam]]]:
        """Get a prompt template from the MCP server.

        Args:
            name: Name of the prompt template

        Returns:
            A callable that accepts keyword arguments and returns a list of BaseMessageParam
        """

        # TODO: wrap this with `llm.prompt` once it's implemented so that they prompts
        # can be run super easily inside of an `llm.context` block.
        async def async_prompt(**kwargs: str) -> list[BaseMessageParam]:
            result = await self._session.get_prompt(name, kwargs)  # type: ignore

            return [
                _utils.convert_prompt_message_to_base_message_params(prompt_message)
                for prompt_message in result.messages
            ]

        return async_prompt

    async def list_tools(self) -> list[type[BaseTool]]:  # pyright: ignore [reportIncompatibleMethodOverride]
        """List all tools available on the MCP server.

        Returns:
            A list of dynamically created `BaseTool` types.
        """
        list_tool_result = await self._session.list_tools()

        converted_tools = []
        for tool in list_tool_result.tools:
            model = _utils.create_tool_from_mcp_tool(tool)
            tool_call_method = _utils.create_tool_call(
                tool.name, self._session.call_tool
            )
            model.call = tool_call_method  # pyright: ignore [reportAttributeAccessIssue]
            converted_tools.append(model)
        return converted_tools


@contextlib.asynccontextmanager
async def sse_client(
    url: str,
    list_roots_callback: ListRootsFnT | None = None,
    read_timeout_seconds: timedelta | None = None,
    sampling_callback: SamplingFnT | None = None,
    session: ClientSession | None = None,
) -> AsyncIterator[MCPClient]:
    async with (
        mcp_sse_client(url) as (read, write),
        ClientSession(
            read,
            write,
            read_timeout_seconds=read_timeout_seconds,
            sampling_callback=sampling_callback,
            list_roots_callback=list_roots_callback,
        ) as session,
    ):
        await session.initialize()
        yield MCPClient(session)


@contextlib.asynccontextmanager
async def stdio_client(
    server_parameters: StdioServerParameters,
    read_stream_exception_handler: Callable[[Exception], None] | None = None,
) -> AsyncIterator[MCPClient]:
    """
    Create a MCPClient instance with the given server parameters and exception handler.

    Args:
        server_parameters:
        read_stream_exception_handler:

    Returns:

    """
    async with (
        mcp_stdio_client(server_parameters) as (read, write),
        _utils.read_stream_exception_filer(
            read, read_stream_exception_handler
        ) as filtered_read,
        ClientSession(filtered_read, write) as session,  # pyright: ignore [reportArgumentType]
    ):
        await session.initialize()
        yield MCPClient(session)



import json
from unittest.mock import MagicMock

from azure.ai.inference.models import (
    AssistantMessage,
    ImageContentItem,
    ImageUrl,
    TextContentItem,
    ToolMessage,
    UserMessage,
)

from mirascope.core import BaseMessageParam
from mirascope.core.azure._utils._message_param_converter import (
    AzureMessageParamConverter,
)
from mirascope.core.base import ImageURLPart
from mirascope.core.base.message_param import TextPart, ToolCallPart, ToolResultPart


def test_convert_no_tool_calls_with_content():
    """
    If there are no tool_calls and content is present, we expect a single BaseMessageParam
    with role="assistant" and content as a string.
    """
    message_param = MagicMock(spec=AssistantMessage)
    message_param.tool_calls = None
    message_param.content = "Hello Azure"

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert result.content == "Hello Azure"


def test_convert_content_list():
    """Test that we can convert a list of content parts properly."""
    message_param = MagicMock(spec=AssistantMessage)
    message_param.tool_calls = None
    message_param.content = [
        TextContentItem(text="Content 1"),
        TextContentItem(text="Content 2"),
        ImageContentItem(image_url=ImageUrl(url="https://example.com/image.jpg")),
    ]

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert result.role == "assistant"
    assert len(result.content) == 3
    assert result.content[0] == TextPart(type="text", text="Content 1")
    assert result.content[1] == TextPart(type="text", text="Content 2")
    assert result.content[2] == ImageURLPart(
        type="image_url", url="https://example.com/image.jpg", detail=None
    )

    message_param = MagicMock(spec=UserMessage)
    message_param.tool_calls = None
    message_param.content = [
        TextContentItem(text="Content 1"),
        TextContentItem(text="Content 2"),
        ImageContentItem(image_url=ImageUrl(url="https://example.com/image.jpg")),
    ]

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert result.role == "user"
    assert len(result.content) == 3
    assert result.content[0] == TextPart(type="text", text="Content 1")
    assert result.content[1] == TextPart(type="text", text="Content 2")
    assert result.content[2] == ImageURLPart(
        type="image_url", url="https://example.com/image.jpg", detail=None
    )


def test_convert_no_tool_calls_no_content():
    """
    If there are no tool_calls and no content, expect a single BaseMessageParam
    with role="assistant" and empty string content.
    """
    message_param = MagicMock(spec=AssistantMessage)
    message_param.tool_calls = None
    message_param.content = None  # or ""

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert result.content == ""


def test_convert_with_single_tool_call():
    """
    If there is a single tool_call, we expect role="tool" and a single ToolCallPart.
    """
    function = MagicMock()
    function.name = "test_tool"
    function.arguments = json.dumps({"param": "value"})

    tool_call = MagicMock()
    tool_call.id = "tool_call_1"
    tool_call.function = function

    message_param = MagicMock(spec=AssistantMessage)
    message_param.tool_calls = [tool_call]
    message_param.content = "Content should be ignored because of tool_calls"

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert len(result.content) == 2
    part = result.content[0]
    assert isinstance(part, ToolCallPart)
    assert part.name == "test_tool"
    assert part.args == {"param": "value"}
    assert part.id == "tool_call_1"


def test_convert_with_multiple_tool_calls():
    """
    If there are multiple tool_calls, expect role="tool" and multiple ToolCallParts.
    """
    # Mock first tool_call
    function1 = MagicMock()
    function1.name = "tool_one"
    function1.arguments = json.dumps({"key1": "val1"})

    tool_call1 = MagicMock()
    tool_call1.id = "tc_1"
    tool_call1.function = function1

    # Mock second tool_call
    function2 = MagicMock()
    function2.name = "tool_two"
    function2.arguments = json.dumps({"key2": "val2"})

    tool_call2 = MagicMock()
    tool_call2.id = "tc_2"
    tool_call2.function = function2

    message_param = MagicMock(spec=AssistantMessage)
    message_param.tool_calls = [tool_call1, tool_call2]
    message_param.content = "Ignored content"

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert len(result.content) == 3

    part1 = result.content[0]
    assert isinstance(part1, ToolCallPart)
    assert part1.name == "tool_one"
    assert part1.args == {"key1": "val1"}
    assert part1.id == "tc_1"

    part2 = result.content[1]
    assert isinstance(part2, ToolCallPart)
    assert part2.name == "tool_two"
    assert part2.args == {"key2": "val2"}
    assert part2.id == "tc_2"


def test_convert_tool_message_param():
    """Tests converting a message parameter with the tool role."""
    message_param = MagicMock(spec=ToolMessage)
    message_param.role = "tool"
    message_param.content = "Message before tool"
    message_param.tool_call_id = "id"

    results = AzureMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "tool"
    assert len(result.content) == 1
    assert result.content[0] == ToolResultPart(
        type="tool_result",
        name="",
        content="Message before tool",
        id="id",
        is_error=False,
    )


def test_base_message_param_to_provider():
    """
    If we have a BaseMessageParam, we should be able to convert it to AzureMessageParam.
    """
    message_param = BaseMessageParam(role="user", content="Hello Azure")
    results = AzureMessageParamConverter.to_provider([message_param])
    assert results == [{"role": "user", "content": "Hello Azure"}]


def test_empty_base_message_param_to_provider():
    """
    If we have an empty BaseMessageParam, we should get an empty list.
    """
    results = AzureMessageParamConverter.to_provider([])
    assert results == []



"""A function generating content to request JSON mode from models without it."""

import json

from pydantic import BaseModel

from ..tool import GenerateJsonSchemaNoTitles


def json_mode_content(tool_type: type[BaseModel] | None) -> str:
    """Returns the content to request JSON mode from models without it."""
    if not tool_type:
        return "\n\nFor your final response, output ONLY a valid JSON dict that adheres to the schema"
    return f"""

For your final response, output ONLY a valid JSON dict (NOT THE SCHEMA) from the content that adheres to this schema:
{json.dumps(tool_type.model_json_schema(schema_generator=GenerateJsonSchemaNoTitles), indent=2)}"""



"""This module contains the `BedrockCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from collections.abc import Sequence
from functools import cached_property
from typing import cast

from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef as SyncConverseResponseTypeDef,
)
from pydantic import SerializeAsAny, SkipValidation, computed_field
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef as AsyncConverseResponseTypeDef,
)

from .. import BaseMessageParam
from ..base import (
    BaseCallResponse,
    transform_tool_outputs,
)
from ..base.types import CostMetadata, FinishReason
from ._call_kwargs import BedrockCallKwargs
from ._types import (
    AssistantMessageTypeDef,
    AsyncMessageTypeDef,
    InternalBedrockMessageParam,
    SyncMessageTypeDef,
    TokenUsageTypeDef,
    ToolResultBlockContentTypeDef,
    ToolResultBlockMessageTypeDef,
    ToolTypeDef,
    ToolUseBlockContentTypeDef,
    UserMessageTypeDef,
)
from ._utils._convert_finish_reason_to_common_finish_reasons import (
    _convert_finish_reasons_to_common_finish_reasons,
)
from ._utils._message_param_converter import BedrockMessageParamConverter
from .call_params import BedrockCallParams
from .dynamic_config import AsyncBedrockDynamicConfig, BedrockDynamicConfig
from .tool import BedrockTool


class BedrockCallResponse(
    BaseCallResponse[
        SyncConverseResponseTypeDef | AsyncConverseResponseTypeDef,
        BedrockTool,
        ToolTypeDef,
        AsyncBedrockDynamicConfig | BedrockDynamicConfig,
        InternalBedrockMessageParam,
        BedrockCallParams,
        UserMessageTypeDef,
        BedrockMessageParamConverter,
    ]
):
    """A convenience wrapper around the Bedrock `ChatCompletion` response.

    When calling the Bedrock API using a function decorated with `bedrock_call`, the
    response will be an `BedrockCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.bedrock import bedrock_call


    @bedrock_call("anthropic.claude-3-haiku-20240307-v1:0")
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    response = recommend_book("fantasy")  # response is an `BedrockCallResponse` instance
    print(response.content)
    ```
    """

    response: SkipValidation[SyncConverseResponseTypeDef | AsyncConverseResponseTypeDef]
    _message_converter: type[BedrockMessageParamConverter] = (
        BedrockMessageParamConverter
    )

    _provider = "bedrock"

    @property
    def message(self) -> SyncMessageTypeDef | AsyncMessageTypeDef | None:
        """Returns the message of the response."""
        message = self.response.get("output", {}).get("message", {})
        if message:
            return cast(SyncMessageTypeDef | AsyncMessageTypeDef, message)
        return None

    @computed_field
    @property
    def content(self) -> str:
        """Returns the content of the chat completion for the 0th choice."""
        if not self.message:
            return ""
        if content := self.message.get("content", []):
            return content[0].get("text", "")
        return ""

    @computed_field
    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [self.response["stopReason"]]

    @computed_field
    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return cast(BedrockCallKwargs, self.call_kwargs)["modelId"]

    @computed_field
    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response["ResponseMetadata"]["RequestId"]

    @property
    def usage(self) -> TokenUsageTypeDef:
        """Returns the usage of the chat completion."""
        return self.response["usage"]

    @computed_field
    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return self.usage["inputTokens"] if self.usage else None

    @property
    def cached_tokens(self) -> int | None:
        """Returns the number of cached tokens."""
        return None

    @computed_field
    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage["outputTokens"] if self.usage else None

    @computed_field
    @cached_property
    def message_param(self) -> SerializeAsAny[AssistantMessageTypeDef]:
        """Returns the assistants's response as a message parameter."""
        message = self.message
        if not message:
            return AssistantMessageTypeDef(role="assistant", content=[])
        return AssistantMessageTypeDef(role="assistant", content=message["content"])

    @cached_property
    def tools(self) -> list[BedrockTool] | None:
        """Returns any available tool calls as their `BedrockTool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
            ValueError: if the model refused to response, in which case the error
                message will be the refusal.
        """
        if not self.message:
            return None
        content = self.message.get("content", [])
        tool_uses = [t for c in content if (t := c.get("toolUse"))]
        if not self.tool_types or not tool_uses:
            return None

        extracted_tools = []
        for tool_use in tool_uses:
            for tool_type in self.tool_types:
                if tool_use["name"] == tool_type._name():
                    extracted_tools.append(
                        tool_type.from_tool_call(
                            cast(ToolUseBlockContentTypeDef, {"toolUse": tool_use})
                        )
                    )
                    break

        return extracted_tools

    @cached_property
    def tool(self) -> BedrockTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
            ValueError: if the model refused to response, in which case the error
                message will be the refusal.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[BedrockTool, str]]
    ) -> list[ToolResultBlockMessageTypeDef]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `ChatCompletionToolMessageParam` parameters.
        """
        return [
            ToolResultBlockMessageTypeDef(
                role="user",
                content=[
                    cast(
                        ToolResultBlockContentTypeDef,
                        {
                            "toolResult": {
                                "content": [{"text": output}],
                                "toolUseId": tool.tool_call["toolUse"]["toolUseId"],  # pyright: ignore [reportOptionalSubscript]
                            }
                        },
                    )
                ],
            )
            for tool, output in tools_and_outputs
        ]

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return _convert_finish_reasons_to_common_finish_reasons(self.finish_reasons)

    @property
    def common_message_param(self) -> BaseMessageParam:
        return BedrockMessageParamConverter.from_provider([self.message_param])[0]

    @property
    def common_user_message_param(self) -> BaseMessageParam | None:
        if not self.user_message_param:
            return None
        return BedrockMessageParamConverter.from_provider([self.user_message_param])[0]  # pyright: ignore [reportArgumentType]

    @computed_field
    @property
    def cost_metadata(self) -> CostMetadata:
        """Get metadata required for cost calculation."""
        return super().cost_metadata



"""Utility for converting `BaseMessageParam` to `ChatMessage`."""

import base64

from mistralai.models import (
    AssistantMessage,
    FunctionCall,
    ImageURL,
    ImageURLChunk,
    SystemMessage,
    TextChunk,
    ToolCall,
    ToolMessage,
    UserMessage,
)

from ...base import BaseMessageParam


def _make_message(
    role: str,
    **kwargs,  # noqa: ANN003
) -> AssistantMessage | SystemMessage | ToolMessage | UserMessage:
    if role == "assistant":
        return AssistantMessage(**kwargs)
    elif role == "system":
        return SystemMessage(**kwargs)
    elif role == "tool":
        return ToolMessage(**kwargs)
    elif role == "user":
        return UserMessage(**kwargs)
    raise ValueError(f"Invalid role: {role}")


def convert_message_params(
    message_params: list[
        BaseMessageParam | AssistantMessage | SystemMessage | ToolMessage | UserMessage
    ],
) -> list[AssistantMessage | SystemMessage | ToolMessage | UserMessage]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance(content := message_param.content, str):
            converted_message_params.append(_make_message(**message_param.model_dump()))
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append(TextChunk(text=part.text))

                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/gif",
                        "image/webp",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. Mistral"
                            " currently only supports JPEG, PNG, GIF, and WebP images."
                        )
                    data = base64.b64encode(part.image).decode("utf-8")
                    converted_content.append(
                        ImageURLChunk(
                            image_url=ImageURL(
                                url=f"data:{part.media_type};base64,{data}",
                                detail=part.detail if part.detail else "auto",
                            )
                        )
                    )
                elif part.type == "image_url":
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": part.url,
                        }
                    )
                elif part.type == "tool_call":
                    converted_message_params.append(
                        AssistantMessage(
                            role="assistant",
                            content=converted_content if converted_content else None,
                            tool_calls=[
                                ToolCall(
                                    function=FunctionCall(
                                        name=part.name,
                                        arguments=part.args,  # pyright: ignore [reportArgumentType]
                                    ),
                                    id=part.id,
                                    type="function",
                                ),
                            ],
                        )
                    )
                    converted_content = []
                elif part.type == "tool_result":
                    if converted_content:
                        converted_message_params.append(
                            _make_message(
                                role=message_param.role, content=converted_content
                            )
                        )
                        converted_content = []
                    converted_message_params.append(
                        ToolMessage(
                            role="tool",
                            content=part.content,
                            tool_call_id=part.id,
                            name=part.name,
                        )
                    )
                else:
                    raise ValueError(
                        "Mistral currently only supports text and image parts. "
                        f"Part provided: {part.type}"
                    )
            if converted_content:
                converted_message_params.append(
                    _make_message(role=message_param.role, content=converted_content)
                )
    return converted_message_params



"""Calculate the cost of a completion using the Bedrock API."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str,
) -> float | None:
    """Calculate the cost of a completion using the Bedrock API."""
    # NOTE: We are currently investigating a dynamic approach to determine costs
    # for Bedrock models. This is needed due to the large number of models
    # available through Bedrock, making it impractical to hardcode all pricing
    # information in this initial implementation.
    return None



"""Cost calculation utilities for LLM API calls."""

from __future__ import annotations

from ..base.types import CostMetadata, Provider
from ._anthropic_calculate_cost import (
    calculate_cost as anthropic_calculate_cost,
)
from ._azure_calculate_cost import calculate_cost as azure_calculate_cost
from ._bedrock_calculate_cost import calculate_cost as bedrock_calculate_cost
from ._cohere_calculate_cost import calculate_cost as cohere_calculate_cost
from ._gemini_calculate_cost import calculate_cost as gemini_calculate_cost
from ._google_calculate_cost import calculate_cost as google_calculate_cost
from ._groq_calculate_cost import calculate_cost as groq_calculate_cost
from ._litellm_calculate_cost import calculate_cost as litellm_calculate_cost
from ._mistral_calculate_cost import calculate_cost as mistral_calculate_cost
from ._openai_calculate_cost import calculate_cost as openai_calculate_cost
from ._vertex_calculate_cost import calculate_cost as vertex_calculate_cost
from ._xai_calculate_cost import calculate_cost as xai_calculate_cost


def calculate_cost(
    provider: Provider,
    model: str,
    metadata: CostMetadata | None = None,
) -> float | None:
    """Calculate the cost for an LLM API call.

    This function routes to the appropriate provider-specific cost calculation function,
    preserving existing behavior while providing a unified interface.

    Args:
        provider: The LLM provider (e.g., "openai", "anthropic")
        model: The model name (e.g., "gpt-4", "claude-3-opus")
        metadata: Additional metadata required for cost calculation

    Returns:
        The calculated cost in USD or None if unable to calculate
    """

    # Initialize empty metadata if none provided
    if metadata is None:
        metadata = CostMetadata()

    # Set default values
    if metadata.cached_tokens is None:
        metadata.cached_tokens = 0

    # Route to provider-specific implementations
    if provider == "openai":
        return openai_calculate_cost(metadata, model)

    elif provider == "anthropic":
        return anthropic_calculate_cost(metadata, model)

    elif provider == "azure":
        return azure_calculate_cost(metadata, model)

    elif provider == "bedrock":
        return bedrock_calculate_cost(metadata, model)

    elif provider == "cohere":
        return cohere_calculate_cost(metadata, model)

    elif provider == "gemini":
        return gemini_calculate_cost(metadata, model)

    elif provider == "google":
        return google_calculate_cost(metadata, model)

    elif provider == "groq":
        return groq_calculate_cost(metadata, model)

    elif provider == "mistral":
        return mistral_calculate_cost(metadata, model)

    elif provider == "vertex":
        return vertex_calculate_cost(metadata, model)

    elif provider == "xai":
        return xai_calculate_cost(metadata, model)

    elif provider == "litellm":
        return litellm_calculate_cost(metadata, model)
    else:
        raise ValueError(f"Unsupported provider: {provider}")



"""This module contains the setup_call function for Cohere tools."""

import inspect
from collections.abc import (
    Awaitable,
    Callable,
)
from typing import (
    Any,
    cast,
    overload,
)

from cohere import (
    AsyncClient,
    Client,
    NonStreamedChatResponse,
)
from cohere.types import ChatMessage, StreamedChatResponse
from pydantic import BaseModel

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import (
    AsyncCreateFn,
    CreateFn,
    get_async_create_fn,
    get_create_fn,
)
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import CohereCallKwargs
from ..call_params import CohereCallParams
from ..dynamic_config import AsyncCohereDynamicConfig, CohereDynamicConfig
from ..tool import CohereTool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params


@overload
def setup_call(
    *,
    model: str,
    client: AsyncClient | None,
    fn: Callable[..., Awaitable[AsyncCohereDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncCohereDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: CohereCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[NonStreamedChatResponse, StreamedChatResponse],
    str,
    list[ChatMessage],
    list[type[CohereTool]] | None,
    CohereCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: Client | None,
    fn: Callable[..., CohereDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: CohereDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: CohereCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[NonStreamedChatResponse, StreamedChatResponse],
    str,
    list[ChatMessage],
    list[type[CohereTool]] | None,
    CohereCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: Client | AsyncClient | None,
    fn: Callable[..., CohereDynamicConfig | Awaitable[AsyncCohereDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: CohereDynamicConfig | AsyncCohereDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: CohereCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[NonStreamedChatResponse, StreamedChatResponse]
    | AsyncCreateFn[NonStreamedChatResponse, StreamedChatResponse],
    str | None,
    list[ChatMessage],
    list[type[CohereTool]] | None,
    CohereCallKwargs,
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        CohereTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(CohereCallKwargs, call_kwargs)
    messages = cast(list[BaseMessageParam | ChatMessage], messages)
    messages = convert_message_params(messages)

    preamble = ""
    if "preamble" in call_kwargs and call_kwargs["preamble"] is not None:
        preamble = call_kwargs.pop("preamble", "") or ""
    if messages[0].role == "SYSTEM":  # pyright: ignore [reportAttributeAccessIssue]
        if preamble:
            preamble += "\n\n"
        preamble += messages.pop(0).message
    if preamble:
        call_kwargs["preamble"] = preamble
    if len(messages) > 1:
        call_kwargs["chat_history"] = messages[:-1]
    if json_mode:
        # Cannot mutate ChatMessage in place
        messages[-1] = ChatMessage(
            role=messages[-1].role,  # pyright: ignore [reportCallIssue, reportAttributeAccessIssue]
            message=messages[-1].message + _utils.json_mode_content(response_model),
            tool_calls=messages[-1].tool_calls,
        )
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
    call_kwargs |= {
        "model": model,
        "message": messages[-1].message,
    }

    if client is None:
        client = AsyncClient() if inspect.iscoroutinefunction(fn) else Client()

    create_or_stream = (
        get_async_create_fn(client.chat, client.chat_stream)
        if isinstance(client, AsyncClient)
        else get_create_fn(client.chat, client.chat_stream)
    )

    return create_or_stream, prompt_template, messages, tool_types, call_kwargs



"""Utility for converting `BaseMessageParam` to `ChatCompletionMessageParam`"""

import base64
import json

from groq.types.chat import (
    ChatCompletionMessageParam,
)

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | ChatCompletionMessageParam],
) -> list[ChatCompletionMessageParam]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance(content := message_param.content, str):
            converted_message_params.append(message_param.model_dump())
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append(part.model_dump())
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/gif",
                        "image/webp",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. Groq"
                            " currently only supports JPEG, PNG, GIF, and WebP images."
                        )
                    data = base64.b64encode(part.image).decode("utf-8")
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:{part.media_type};base64,{data}",
                                "detail": part.detail if part.detail else "auto",
                            },
                        }
                    )
                elif part.type == "image_url":
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": part.url,
                            },
                        }
                    )
                elif part.type == "tool_call":
                    converted_message_param = {
                        "role": "assistant",
                        "tool_calls": [
                            {
                                "type": "function",
                                "id": part.id,
                                "function": {
                                    "name": part.name,
                                    "arguments": json.dumps(part.args),
                                },
                            }
                        ],
                    }
                    if converted_content:
                        if len(converted_content) == 1:
                            if converted_content[0]["type"] == "text":
                                converted_message_param["content"] = converted_content[
                                    0
                                ]["text"]
                        else:
                            converted_message_params.append(
                                {
                                    "role": message_param.role,
                                    "content": converted_content,
                                }
                            )
                        converted_content = []
                    converted_message_params.append(converted_message_param)

                elif part.type == "tool_result":
                    if converted_content:
                        converted_message_params.append(
                            {
                                "role": message_param.role,
                                "content": converted_content,
                            }
                        )
                        converted_content = []
                    converted_message_params.append(
                        {
                            "role": "tool",
                            "content": part.content,
                            "tool_call_id": part.id,
                        }
                    )
                else:
                    raise ValueError(
                        "Groq currently only supports text and image parts. "
                        f"Part provided: {part.type}"
                    )
            if converted_content:
                converted_message_params.append(
                    {"role": message_param.role, "content": converted_content}
                )
    return converted_message_params



"""This module defines the function return type for functions as LLM calls."""

from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.aio import ChatCompletionsClient as AsyncChatCompletionsClient
from azure.ai.inference.models import ChatRequestMessage

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import AzureCallParams

AsyncAzureDynamicConfig = BaseDynamicConfig[
    ChatRequestMessage | BaseMessageParam, AzureCallParams, AsyncChatCompletionsClient
]
AzureDynamicConfig = BaseDynamicConfig[
    ChatRequestMessage | BaseMessageParam, AzureCallParams, ChatCompletionsClient
]
"""The function return type for functions wrapped with the `azure_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.azure import AzureDynamicConfig, azure_call


@azure_call("gpt-4o-mini")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> AzureDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""The `BasePrompt` class for better prompt engineering."""

from collections.abc import AsyncIterable, Awaitable, Callable, Iterable
from functools import reduce, wraps
from textwrap import dedent
from typing import (
    Any,
    ClassVar,
    ParamSpec,
    Protocol,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ._utils import (
    BaseType,
    MessagesDecorator,
    fn_is_async,
    format_template,
    get_fn_args,
    get_metadata,
    get_prompt_template,
    messages_decorator,
    parse_prompt_messages,
)
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .message_param import BaseMessageParam
from .metadata import Metadata

_P = ParamSpec("_P")
_R = TypeVar("_R")
_T = TypeVar("_T", bound=Callable)
_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseStreamT = TypeVar("_BaseStreamT")
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)

SUPPORTED_MESSAGE_ROLES = ["system", "user", "assistant"]


class BasePrompt(BaseModel):
    """The base class for engineering prompts.

    This class is implemented as the base for all prompting needs. It is intended to
    work across various providers by providing a common prompt interface.

    Example:

    ```python
    from mirascope.core import BasePrompt, metadata, prompt_template

    @prompt_template("Recommend a {genre} book")
    @metadata({"tags": {"version:0001", "books"}})
    class BookRecommendationPrompt(BasePrompt):
        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")

    print(prompt)
    # > Recommend a fantasy book

    print(prompt.message_params())
    # > [BaseMessageParam(role="user", content="Recommend a fantasy book")]

    print(prompt.dump()["metadata"])
    # > {"metadata": {"version:0001", "books"}}
    ```
    """

    prompt_template: ClassVar[str]

    def __str__(self) -> str:
        """Returns the formatted template."""
        prompt_template = get_prompt_template(self)
        prompt_template = (
            prompt_template.replace(":images}", "}")
            .replace(":image}", "}")
            .replace(":audios}", "}")
            .replace(":audio}", "}")
            .replace(":documents}", "}")
            .replace(":document}", "}")
            .replace(":texts}", "}")
            .replace(":text", "")
        )
        return format_template(prompt_template, self.model_dump())

    def message_params(self) -> list[BaseMessageParam]:
        """Returns the list of parsed message parameters."""
        return parse_prompt_messages(
            roles=SUPPORTED_MESSAGE_ROLES,
            template=get_prompt_template(self),
            attrs={field: getattr(self, field) for field in self.model_fields},
        )

    def dynamic_config(self) -> BaseDynamicConfig:
        """Returns the dynamic config of the prompt."""
        return None

    def dump(self) -> dict[str, Any]:
        """Dumps the contents of the prompt into a dictionary."""
        return {
            "metadata": get_metadata(self, None),
            "prompt": str(self),
            "template": get_prompt_template(self),
            "inputs": self.model_dump(),
        }

    @overload
    def run(
        self,
        call_decorator: Callable[
            [Callable[..., BaseDynamicConfig]], Callable[..., _BaseCallResponseT]
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> _BaseCallResponseT: ...

    @overload
    def run(
        self,
        call_decorator: Callable[
            [Callable[..., BaseDynamicConfig]], Callable[..., _BaseStreamT]
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> _BaseStreamT: ...

    @overload
    def run(
        self,
        call_decorator: Callable[
            [Callable[..., BaseDynamicConfig]], Callable[..., _ResponseModelT]
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> _ResponseModelT: ...

    @overload
    def run(
        self,
        call_decorator: Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., Iterable[_ResponseModelT]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> Iterable[_ResponseModelT]: ...

    def run(
        self,
        call_decorator: Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., _BaseCallResponseT],
        ]
        | Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., _BaseStreamT],
        ]
        | Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., _ResponseModelT],
        ]
        | Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., Iterable[_ResponseModelT]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> (
        _BaseCallResponseT | _BaseStreamT | _ResponseModelT | Iterable[_ResponseModelT]
    ):
        """Returns the response of calling the API of the provided decorator.

        Example:

        ```python
        from mirascope.core import BasePrompt, openai, prompt_template


        @prompt_template("Recommend a {genre} book")
        class BookRecommendationPrompt(BasePrompt):
            genre: str


        prompt = BookRecommendationPrompt(genre="fantasy")
        response = prompt.run(openai.call("gpt-4o-mini"))
        print(response.content)
        ```
        """
        kwargs = {field: getattr(self, field) for field in self.model_fields}
        args_str = ", ".join(kwargs.keys())
        namespace, fn_name = {}, self.__class__.__name__
        exec(f"def {fn_name}({args_str}): ...", namespace)
        return reduce(
            lambda res, f: f(res),  # pyright: ignore [reportArgumentType, reportCallIssue]
            [
                metadata(get_metadata(self, self.dynamic_config())),
                prompt_template(get_prompt_template(self)),
                call_decorator,
                *additional_decorators,
            ],
            namespace[fn_name],
        )(**kwargs)

    @overload
    def run_async(
        self,
        call_decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseCallResponseT]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> Awaitable[_BaseCallResponseT]: ...

    @overload
    def run_async(
        self,
        call_decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseStreamT]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> Awaitable[_BaseStreamT]: ...

    @overload
    def run_async(
        self,
        call_decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_ResponseModelT]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> Awaitable[_ResponseModelT]: ...

    @overload
    def run_async(
        self,
        call_decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[AsyncIterable[_ResponseModelT]]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> Awaitable[AsyncIterable[_ResponseModelT]]: ...

    def run_async(
        self,
        call_decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseCallResponseT]],
        ]
        | Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseStreamT]],
        ]
        | Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_ResponseModelT]],
        ]
        | Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[AsyncIterable[_ResponseModelT]]],
        ],
        *additional_decorators: Callable[[_T], _T],
    ) -> (
        Awaitable[_BaseCallResponseT]
        | Awaitable[_BaseStreamT]
        | Awaitable[_ResponseModelT]
        | Awaitable[AsyncIterable[_ResponseModelT]]
    ):
        """Returns the response of calling the API of the provided decorator.

        Example:

        ```python
        import asyncio

        from mirascope.core import BasePrompt, openai, prompt_template


        @prompt_template("Recommend a {genre} book")
        class BookRecommendationPrompt(BasePrompt):
            genre: str


        async def run():
            prompt = BookRecommendationPrompt(genre="fantasy")
            response = await prompt.run_async(openai.call("gpt-4o-mini"))
            print(response.content)


        asyncio.run(run())
        ```
        """
        kwargs = {field: getattr(self, field) for field in self.model_fields}
        args_str = ", ".join(kwargs.keys())
        namespace, fn_name = {}, self.__class__.__name__
        exec(f"async def {fn_name}({args_str}): ...", namespace)
        return reduce(
            lambda res, f: f(res),  # pyright: ignore [reportArgumentType, reportCallIssue]
            [
                metadata(get_metadata(self, self.dynamic_config())),
                prompt_template(get_prompt_template(self)),
                call_decorator,
                *additional_decorators,
            ],
            namespace[fn_name],
        )(**kwargs)


_BasePromptT = TypeVar("_BasePromptT", bound=BasePrompt)


class PromptDecorator(Protocol):
    @overload
    def __call__(
        self, prompt: Callable[_P, BaseDynamicConfig]
    ) -> Callable[_P, list[BaseMessageParam]]: ...

    @overload
    def __call__(
        self, prompt: Callable[_P, Awaitable[BaseDynamicConfig]]
    ) -> Callable[_P, Awaitable[list[BaseMessageParam]]]: ...

    @overload
    def __call__(self, prompt: type[_BasePromptT]) -> type[_BasePromptT]: ...

    def __call__(
        self,
        prompt: type[_BasePromptT]
        | Callable[_P, BaseDynamicConfig]
        | Callable[_P, Awaitable[BaseDynamicConfig]],
    ) -> (
        Callable[_P, list[BaseMessageParam]]
        | Callable[_P, Awaitable[list[BaseMessageParam]]]
        | type[_BasePromptT]
    ): ...


@overload
def prompt_template(template: str) -> PromptDecorator: ...


@overload
def prompt_template(template: None = None) -> MessagesDecorator: ...


def prompt_template(
    template: str | None = None,
) -> PromptDecorator | MessagesDecorator:
    """A decorator for setting the `prompt_template` of a `BasePrompt` or `call`.

    usage docs: learn/prompts.md#prompt-templates-messages

    Example:

    ```python
    from mirascope.core import openai, prompt_template


    @prompt_template()
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    print(recommend_book("fantasy"))
    # Output: [BaseMessageParam(role='user', content='Recommend a fantasy book')]
    ```

    Returns:
        decorator (Callable): The decorator function that turns the decorated function
            into a prompt template.
    """

    if template is None:
        # For @prompt_template() case
        decorator = messages_decorator()
        decorator.__mirascope_prompt_template__ = True  # pyright: ignore [reportAttributeAccessIssue]
        return decorator
    template = dedent(template).strip()

    @overload
    def inner(
        prompt: Callable[_P, BaseDynamicConfig],
    ) -> Callable[_P, list[BaseMessageParam]]: ...

    @overload
    def inner(
        prompt: Callable[_P, Awaitable[BaseDynamicConfig]],
    ) -> Callable[_P, Awaitable[list[BaseMessageParam]]]: ...

    @overload
    def inner(prompt: type[_BasePromptT]) -> type[_BasePromptT]: ...

    def inner(
        prompt: type[_BasePromptT]
        | Callable[_P, BaseDynamicConfig]
        | Callable[_P, Awaitable[BaseDynamicConfig]],
    ) -> (
        Callable[_P, list[BaseMessageParam]]
        | Callable[_P, Awaitable[list[BaseMessageParam]]]
        | type[_BasePromptT]
    ):
        """Updates the `prompt_template` class attribute to the given value."""
        prompt._prompt_template = template  # pyright: ignore [reportAttributeAccessIssue,reportFunctionMemberAccess]

        if isinstance(prompt, type):
            return prompt

        if fn_is_async(prompt):

            @wraps(prompt)
            async def get_base_message_params_async(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> list[BaseMessageParam]:
                return parse_prompt_messages(
                    roles=SUPPORTED_MESSAGE_ROLES,
                    template=template,
                    attrs=get_fn_args(prompt, args, kwargs),
                    dynamic_config=await prompt(*args, **kwargs),
                )

            get_base_message_params_async._original_fn = prompt  # pyright: ignore [reportAttributeAccessIssue,reportFunctionMemberAccess]
            return get_base_message_params_async

        else:

            @wraps(prompt)
            def get_base_message_params(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> list[BaseMessageParam]:
                return parse_prompt_messages(
                    roles=SUPPORTED_MESSAGE_ROLES,
                    template=template,
                    attrs=get_fn_args(prompt, args, kwargs),
                    dynamic_config=prompt(*args, **kwargs),
                )

            get_base_message_params._original_fn = prompt  # pyright: ignore [reportAttributeAccessIssue,reportFunctionMemberAccess]
            return get_base_message_params

    inner.__mirascope_prompt_template__ = True  # pyright: ignore [reportFunctionMemberAccess]
    return inner


class MetadataDecorator(Protocol):
    @overload
    def __call__(self, prompt: type[_BasePromptT]) -> type[_BasePromptT]: ...

    @overload
    def __call__(self, prompt: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def __call__(
        self,
        prompt: type[_BasePromptT] | Callable[_P, _R],
    ) -> type[_BasePromptT] | Callable[_P, _R]: ...


def metadata(metadata: Metadata) -> MetadataDecorator:
    """A decorator for adding metadata to a `BasePrompt` or `call`.

    Adding this decorator to a `BasePrompt` or `call` updates the `metadata` annotation
    to the given value. This is useful for adding metadata to a `BasePrompt` or `call`
    that can be used for logging or filtering.

    Example:

    ```python
    from mirascope.core import metadata, openai, prompt_template


    @openai.call("gpt-4o-mini")
    @prompt_template("Recommend a {genre} book")
    @metadata({"tags": {"version:0001", "books"}})
    def recommend_book(genre: str):
        ...


    response = recommend_book("fantasy")
    print(response.metadata)
    ```

    Returns:
        decorator (Callable): The decorator function that updates the `_metadata`
            attribute of the decorated input prompt or call.
    """

    @overload
    def inner(prompt: type[_BasePromptT]) -> type[_BasePromptT]: ...

    @overload
    def inner(prompt: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def inner(
        prompt: type[_BasePromptT] | Callable[_P, _R],
    ) -> type[_BasePromptT] | Callable[_P, _R]:
        """Updates the `metadata` class attribute to the given value."""
        prompt._metadata = metadata  # pyright: ignore [reportAttributeAccessIssue,reportFunctionMemberAccess]
        return prompt

    return inner



"""Calculate the cost of a completion using the OpenAI API."""

from ..base.types import CostMetadata

# Constants for image token calculation
LOW_DETAIL_IMAGE_TOKENS = 85
HIGH_DETAIL_TILE_TOKENS = 170
HIGH_DETAIL_BASE_TOKENS = 85
TILE_SIZE = 512


def _calculate_image_tokens(metadata: CostMetadata) -> int | None:
    """Calculate tokens used by images based on their size and detail level.

    https://platform.openai.com/docs/guides/vision
    Following OpenAI's pricing structure:
    - Low detail: 85 tokens per image
    - High detail: 85 tokens base + 170 tokens per 512px tile
      (after scaling to fit within 2048x2048 and making shortest side 768px)
    """
    if not metadata.images or not metadata.images:
        return 0

    total_image_tokens = 0

    for img in metadata.images:
        if not img.width or not img.height:
            continue

        # If image already has precalculated tokens, use those
        if img.tokens is not None:
            total_image_tokens += img.tokens
            continue

        if img.detail is not None and img.detail != "auto":
            detail = img.detail
        else:
            # Default to high detail for auto
            # We can't determine detail level from image alone
            detail = "high"
        if detail == "low":
            # Low detail is a fixed cost regardless of size
            total_image_tokens += LOW_DETAIL_IMAGE_TOKENS
        else:
            # High detail calculation

            # Scale to fit within 2048x2048 square
            width, height = img.width, img.height
            if width > 2048 or height > 2048:
                aspect_ratio = width / height
                if width > height:
                    width = 2048
                    height = int(width / aspect_ratio)
                else:
                    height = 2048
                    width = int(height * aspect_ratio)

            # Scale so shortest side is 768px
            if min(width, height) > 768:
                if width < height:
                    scale_factor = 768 / width
                    width = 768
                    height = int(height * scale_factor)
                else:
                    scale_factor = 768 / height
                    height = 768
                    width = int(width * scale_factor)

            # Calculate number of 512px tiles needed
            tiles_x = (width + TILE_SIZE - 1) // TILE_SIZE
            tiles_y = (height + TILE_SIZE - 1) // TILE_SIZE
            num_tiles = tiles_x * tiles_y

            # Calculate token cost
            image_tokens = (
                HIGH_DETAIL_TILE_TOKENS * num_tiles
            ) + HIGH_DETAIL_BASE_TOKENS
            total_image_tokens += image_tokens

    return total_image_tokens


def calculate_cost(
    metadata: CostMetadata,
    model: str = "gpt-3.5-turbo-16k",
) -> float | None:
    """Calculate the cost of a completion using the OpenAI API.

    https://openai.com/pricing

    Model                                     Input               Cached               Output
    gpt-4.5-preview                           $75.00 / 1M tokens  $37.50 / 1M tokens   $150.00 / 1M tokens
    gpt-4.5-preview-2025-02-27                $75.00 / 1M tokens  $37.50 / 1M tokens   $150.00 / 1M tokens
    gpt-4o                                    $2.50  / 1M tokens  $1.25 / 1M tokens    $10.00 / 1M tokens
    gpt-4o-2024-11-20                         $2.50  / 1M tokens  $1.25 / 1M tokens    $10.00 / 1M tokens
    gpt-4o-2024-08-06                         $2.50  / 1M tokens  $1.25 / 1M tokens    $10.00 / 1M tokens
    gpt-4o-2024-05-13                         $5.00  / 1M tokens                       $15.00 / 1M tokens
    gpt-4o-audio-preview                      $2.50  / 1M tokens  $1.25 / 1M tokens    $10.00 / 1M tokens
    gpt-4o-audio-preview-2024-12-17           $2.50  / 1M tokens  $1.25 / 1M tokens    $10.00 / 1M tokens
    gpt-4o-audio-preview-2024-10-01           $2.50  / 1M tokens  $1.25 / 1M tokens    $10.00 / 1M tokens
    gpt-4o-realtime-preview                   $5.00  / 1M tokens  $2.50 / 1M tokens    $20.00 / 1M tokens
    gpt-4o-realtime-preview-2024-12-17        $5.00  / 1M tokens  $2.50 / 1M tokens    $20.00 / 1M tokens
    gpt-4o-realtime-preview-2024-10-01        $5.00  / 1M tokens  $2.50 / 1M tokens    $20.00 / 1M tokens
    gpt-4o-mini                               $0.150 / 1M tokens  $0.075 / 1M tokens   $0.600 / 1M tokens
    gpt-4o-mini-2024-07-18                    $0.150 / 1M tokens  $0.075 / 1M tokens   $0.600 / 1M tokens
    gpt-4o-mini-audio-preview                 $0.150 / 1M tokens  $0.075 / 1M tokens   $0.600 / 1M tokens
    gpt-4o-mini-audio-preview-2024-12-17      $0.150 / 1M tokens  $0.075 / 1M tokens   $0.600 / 1M tokens
    gpt-4o-mini-realtime-preview              $0.60  / 1M tokens  $0.30 / 1M tokens    $2.40  / 1M tokens
    gpt-4o-mini-realtime-preview-2024-12-17   $0.60  / 1M tokens  $0.30 / 1M tokens    $2.40  / 1M tokens
    o1                                        $15.00 / 1M tokens  $7.50 / 1M tokens    $60.00 / 1M tokens
    o1-2024-12-17                             $15.00 / 1M tokens  $7.50 / 1M tokens    $60.00 / 1M tokens
    o1-preview                                $15.00 / 1M tokens  $7.50 / 1M tokens    $60.00 / 1M tokens
    o1-preview-2024-09-12                     $15.00 / 1M tokens  $7.50 / 1M tokens    $60.00 / 1M tokens
    o3-mini                                   $1.10  / 1M tokens  $0.55 / 1M tokens    $4.40  / 1M tokens
    o3-mini-2025-01-31                        $1.10  / 1M tokens  $0.55 / 1M tokens    $4.40  / 1M tokens
    o1-mini                                   $1.10  / 1M tokens  $0.55 / 1M tokens    $4.40  / 1M tokens
    o1-mini-2024-09-12                        $1.10  / 1M tokens  $0.55 / 1M tokens    $4.40  / 1M tokens
    chatgpt-4o-latest                         $5.00  / 1M tokens                       $15.00 / 1M tokens
    gpt-4-turbo                               $10.00 / 1M tokens                       $30.00 / 1M tokens
    gpt-4-turbo-2024-04-09                    $10.00 / 1M tokens                       $30.00 / 1M tokens
    gpt-3.5-turbo-0125                        $0.50  / 1M tokens                       $1.50  / 1M tokens
    gpt-3.5-turbo-1106                        $1.00  / 1M tokens                       $2.00  / 1M tokens
    gpt-4-0125-preview                        $10.00 / 1M tokens                       $30.00 / 1M tokens
    gpt-4-1106-preview                        $10.00 / 1M tokens                       $30.00 / 1M tokens
    gpt-4-vision-preview                      $10.00 / 1M tokens                       $30.00 / 1M tokens
    gpt-4                                     $30.00 / 1M tokens                       $60.00 / 1M tokens
    gpt-4-32k                                 $60.00 / 1M tokens                       $120.00 / 1M tokens
    text-embedding-3-small                    $0.02  / 1M tokens
    text-embedding-3-large                    $0.13  / 1M tokens
    text-embedding-ada-002                    $0.10  / 1M tokens
    """
    pricing = {
        "gpt-4.5-preview": {
            "prompt": 0.000_075,
            "cached": 0.000_037_5,
            "completion": 0.000_15,
            "batch_prompt": 0.000_037_5,
            "batch_completion": 0.000_075,
        },
        "gpt-4.5-preview-2025-02-27": {
            "prompt": 0.000_075,
            "cached": 0.000_037_5,
            "completion": 0.000_15,
        },
        "gpt-4o": {
            "prompt": 0.000_002_5,
            "cached": 0.000_001_25,
            "completion": 0.000_01,
            "batch_prompt": 0.000_001_25,
            "batch_completion": 0.000_005,
        },
        "gpt-4o-2024-11-20": {
            "prompt": 0.000_002_5,
            "cached": 0.000_001_25,
            "completion": 0.000_01,
        },
        "gpt-4o-2024-08-06": {
            "prompt": 0.000_002_5,
            "cached": 0.000_001_25,
            "completion": 0.000_01,
        },
        "gpt-4o-2024-05-13": {
            "prompt": 0.000_005,
            "cached": 0.000_002_5,
            "completion": 0.000_015,
        },
        "gpt-4o-audio-preview": {
            "prompt": 0.000_002_5,
            "cached": 0.000_001_25,
            "completion": 0.000_01,
        },
        "gpt-4o-audio-preview-2024-12-17": {
            "prompt": 0.000_002_5,
            "cached": 0.000_001_25,
            "completion": 0.000_01,
        },
        "gpt-4o-audio-preview-2024-10-01": {
            "prompt": 0.000_002_5,
            "cached": 0.000_001_25,
            "completion": 0.000_01,
        },
        "gpt-4o-realtime-preview": {
            "prompt": 0.000_005,
            "cached": 0.000_002_5,
            "completion": 0.000_02,
        },
        "gpt-4o-realtime-preview-2024-12-17": {
            "prompt": 0.000_005,
            "cached": 0.000_002_5,
            "completion": 0.000_02,
        },
        "gpt-4o-realtime-preview-2024-10-01": {
            "prompt": 0.000_005,
            "cached": 0.000_002_5,
            "completion": 0.000_02,
        },
        "gpt-4o-mini": {
            "prompt": 0.000_000_15,
            "cached": 0.000_000_075,
            "completion": 0.000_000_6,
        },
        "gpt-4o-mini-2024-07-18": {
            "prompt": 0.000_000_15,
            "cached": 0.000_000_075,
            "completion": 0.000_000_6,
        },
        "gpt-4o-mini-audio-preview": {
            "prompt": 0.000_000_15,
            "cached": 0.000_000_075,
            "completion": 0.000_000_6,
        },
        "gpt-4o-mini-audio-preview-2024-12-17": {
            "prompt": 0.000_000_15,
            "cached": 0.000_000_075,
            "completion": 0.000_000_6,
        },
        "gpt-4o-mini-realtime-preview": {
            "prompt": 0.000_000_6,
            "cached": 0.000_000_3,
            "completion": 0.000_002_4,
        },
        "gpt-4o-mini-realtime-preview-2024-12-17": {
            "prompt": 0.000_000_6,
            "cached": 0.000_000_3,
            "completion": 0.000_002_4,
        },
        "o1": {
            "prompt": 0.000_015,
            "cached": 0.000_007_5,
            "completion": 0.000_06,
        },
        "o1-2024-12-17": {
            "prompt": 0.000_015,
            "cached": 0.000_007_5,
            "completion": 0.000_06,
        },
        "o1-preview": {
            "prompt": 0.000_015,
            "cached": 0.000_007_5,
            "completion": 0.000_06,
        },
        "o1-preview-2024-09-12": {
            "prompt": 0.000_015,
            "cached": 0.000_007_5,
            "completion": 0.000_06,
        },
        "o3-mini": {
            "prompt": 0.000_001_1,
            "cached": 0.000_000_55,
            "completion": 0.000_004_4,
        },
        "o3-mini-2025-01-31": {
            "prompt": 0.000_001_1,
            "cached": 0.000_000_55,
            "completion": 0.000_004_4,
        },
        "o1-mini": {
            "prompt": 0.000_001_1,
            "cached": 0.000_000_55,
            "completion": 0.000_004_4,
        },
        "o1-mini-2024-09-12": {
            "prompt": 0.000_001_1,
            "cached": 0.000_000_55,
            "completion": 0.000_004_4,
        },
        "chatgpt-4o-latest": {
            "prompt": 0.000_005,
            "cached": 0,
            "completion": 0.000_015,
        },
        "gpt-4-turbo": {
            "prompt": 0.000_01,
            "cached": 0,
            "completion": 0.000_03,
        },
        "gpt-4-turbo-2024-04-09": {
            "prompt": 0.000_01,
            "cached": 0,
            "completion": 0.000_03,
        },
        "gpt-3.5-turbo-0125": {
            "prompt": 0.000_000_5,
            "cached": 0,
            "completion": 0.000_001_5,
        },
        "gpt-3.5-turbo-1106": {
            "prompt": 0.000_001,
            "cached": 0,
            "completion": 0.000_002,
        },
        "gpt-4-0125-preview": {
            "prompt": 0.000_01,
            "cached": 0,
            "completion": 0.000_03,
        },
        "gpt-4-1106-preview": {
            "prompt": 0.000_01,
            "cached": 0,
            "completion": 0.000_03,
        },
        "gpt-4-vision-preview": {
            "prompt": 0.000_01,
            "cached": 0,
            "completion": 0.000_03,
        },
        "gpt-4": {
            "prompt": 0.000_03,
            "cached": 0,
            "completion": 0.000_06,
        },
        "gpt-4-32k": {
            "prompt": 0.000_06,
            "cached": 0,
            "completion": 0.000_12,
        },
        "gpt-3.5-turbo-4k": {
            "prompt": 0.000_015,
            "cached": 0,
            "completion": 0.000_02,
        },
        "gpt-3.5-turbo-16k": {
            "prompt": 0.000_003,
            "cached": 0,
            "completion": 0.000_004,
        },
        "gpt-4-8k": {
            "prompt": 0.000_03,
            "cached": 0,
            "completion": 0.000_06,
        },
        "text-embedding-3-small": {
            "prompt": 0.000_000_02,
            "cached": 0,
            "completion": 0,
            "batch_prompt": 0.000_000_01,
        },
        "text-embedding-ada-002": {
            "prompt": 0.000_000_1,
            "cached": 0,
            "completion": 0,
            "batch_prompt": 0.000_000_05,
        },
        "text-embedding-3-large": {
            "prompt": 0.000_000_13,
            "cached": 0,
            "completion": 0,
            "batch_prompt": 0.000_000_065,
        },
    }

    # Audio pricing for audio models (per-minute rates in dollars)

    if metadata.cost is not None:
        return metadata.cost

    # Audio input/output costs
    # ChatCompletion.usage has brake down of audio input and output.
    # The total cost already includes the audio input/output cost.

    # Initialize cached tokens if not provided
    if metadata.cached_tokens is None:
        metadata.cached_tokens = 0

    # Try to get model pricing
    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    image_tokens = _calculate_image_tokens(metadata) or 0

    input_tokens = (metadata.input_tokens or 0) + image_tokens

    # Calculate costs for each component
    prompt_cost = input_tokens * model_pricing["prompt"]
    cached_cost = metadata.cached_tokens * model_pricing["cached"]
    completion_cost = (metadata.output_tokens or 0) * model_pricing["completion"]

    # Special handling for embedding models (only input tokens matter)
    if "embedding" in model:
        total_cost = prompt_cost
    else:
        total_cost = prompt_cost + cached_cost + completion_cost

    # Apply batch discounts if applicable
    if metadata.batch_mode:
        # Based on the OpenAI pricing table, batch mode typically provides
        # approximately 50% discount for both input and output tokens
        if "embedding" in model.lower():
            # Embedding models have specific batch pricing
            if model == "text-embedding-3-small":
                prompt_cost = (
                    input_tokens * 0.000_000_01
                )  # $0.01 per 1M tokens in batch mode
            elif model == "text-embedding-3-large":
                prompt_cost = (
                    input_tokens * 0.000_000_065
                )  # $0.065 per 1M tokens in batch mode
            elif model == "text-embedding-ada-002":
                prompt_cost = (
                    input_tokens * 0.000_000_05
                )  # $0.05 per 1M tokens in batch mode
        else:
            # For LLM models, typically 50% discount
            prompt_cost *= 0.5
            cached_cost *= 0.5
            completion_cost *= 0.5

        # Recalculate total cost with batch pricing
        if "embedding" in model:
            total_cost = prompt_cost
        else:
            total_cost = prompt_cost + cached_cost + completion_cost

    return total_cost



from unittest.mock import AsyncMock, MagicMock, patch

from pydantic import BaseModel
from pytest import mark

from mirascope.tools.web._duckduckgo import (
    AsyncDuckDuckGoSearch,
    DuckDuckGoSearch,
    DuckDuckGoSearchConfig,
)


class SearchSchema(BaseModel):
    """Schema for search results"""

    urls: list[str]


# Config Tests
def test_search_config_defaults():
    """Test default configuration values"""
    config = DuckDuckGoSearchConfig()
    assert config.max_results_per_query == 2


def test_search_config_custom():
    """Test setting custom configuration values"""
    config = DuckDuckGoSearchConfig(max_results_per_query=5)
    assert config.max_results_per_query == 5


def test_search_config_from_env():
    """Test loading configuration from environment variables"""
    config = DuckDuckGoSearchConfig.from_env()
    assert isinstance(config, DuckDuckGoSearchConfig)
    assert config.max_results_per_query == 2


# Search Functionality Tests
@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_single_query_success(mock_ddgs):
    """Test successful execution of a single search query"""
    mock_results = [
        {
            "title": "Example Site 1",
            "href": "https://example1.com",
            "body": "This is the first example snippet",
        },
        {
            "title": "Example Site 2",
            "href": "https://example2.com",
            "body": "This is the second example snippet",
        },
    ]
    mock_ddgs_instance = MagicMock()
    mock_ddgs_instance.text.return_value = mock_results
    mock_ddgs.return_value = mock_ddgs_instance

    search = DuckDuckGoSearch(queries=["test query"])
    result = search.call()

    assert "Example Site 1" in result
    assert "https://example1.com" in result
    assert "This is the first example snippet" in result
    mock_ddgs_instance.text.assert_called_once_with(
        "test query", max_results=search._get_config().max_results_per_query
    )


@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_multiple_queries_success(mock_ddgs):
    """Test successful execution of multiple search queries"""
    mock_results = [
        {
            "title": "Multiple Query Result 1",
            "href": "https://example1.com",
            "body": "First query result snippet",
        },
        {
            "title": "Multiple Query Result 2",
            "href": "https://example2.com",
            "body": "Second query result snippet",
        },
    ]
    mock_ddgs_instance = MagicMock()
    mock_ddgs_instance.text.return_value = mock_results
    mock_ddgs.return_value = mock_ddgs_instance

    search = DuckDuckGoSearch(queries=["query1", "query2"])
    result = search.call()

    assert "Multiple Query Result 1" in result
    assert "https://example1.com" in result
    assert "First query result snippet" in result
    assert mock_ddgs_instance.text.call_count == 2


@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_empty_results(mock_ddgs):
    """Test handling of empty search results"""
    mock_ddgs_instance = MagicMock()
    mock_ddgs_instance.text.return_value = []
    mock_ddgs.return_value = mock_ddgs_instance

    search = DuckDuckGoSearch(queries=["test query"])
    result = search.call()

    assert result == ""


# Error Handling Tests
@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_ddgs_initialization_error(mock_ddgs):
    """Test error handling during DDGS initialization"""
    mock_ddgs.side_effect = Exception("DDGS initialization failed")

    search = DuckDuckGoSearch(queries=["test query"])
    result = search.call()

    assert "<class 'Exception'>: Failed to search the web for text" in result


@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_search_execution_error(mock_ddgs):
    """Test error handling during search execution"""
    mock_results = [
        {
            "title": "Error Test",
            "href": None,
            "body": "Test snippet",
        }
    ]
    mock_ddgs_instance = MagicMock()
    mock_ddgs_instance.text.return_value = mock_results
    mock_ddgs.return_value = mock_ddgs_instance

    search = DuckDuckGoSearch(queries=["test query"])
    result = search.call()

    assert "Title: Error Test\nURL: None\nSnippet: Test snippet" in result


@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_url_parsing_error(mock_ddgs):
    """Test error handling during URL parsing"""
    mock_results = [
        {
            "title": "Invalid URL Test",
            "href": None,
            "body": "Test snippet with invalid URL",
        },
        {
            "title": "Valid URL Test",
            "href": "https://example.com",
            "body": "Test snippet with valid URL",
        },
    ]
    mock_ddgs_instance = MagicMock()
    mock_ddgs_instance.text.return_value = mock_results
    mock_ddgs.return_value = mock_ddgs_instance

    search = DuckDuckGoSearch(queries=["test query"])
    result = search.call()

    assert (
        "Title: Invalid URL Test\n"
        "URL: None\n"
        "Snippet: Test snippet with invalid URL\n"
        "\n"
        "Title: Valid URL Test\n"
        "URL: https://example.com\n"
        "Snippet: Test snippet with valid URL"
    ) in result


# Custom Configuration Tests
@patch("mirascope.tools.web._duckduckgo.DDGS")
def test_custom_max_results(mock_ddgs):
    """Test using custom maximum results configuration"""

    # Create new search class with custom config
    CustomSearch = DuckDuckGoSearch.from_config(
        DuckDuckGoSearchConfig(max_results_per_query=5)
    )

    mock_ddgs_instance = MagicMock()
    mock_ddgs_instance.text.return_value = []
    mock_ddgs.return_value = mock_ddgs_instance

    search = CustomSearch(queries=["test query"])  # pyright: ignore [reportCallIssue]
    search.call()

    mock_ddgs_instance.text.assert_called_once_with("test query", max_results=5)


# Integration Tests
def test_integration_with_search_tool_base():
    """Test integration with search tool base class"""
    search = DuckDuckGoSearch(queries=["test"])
    description = search.usage_description()

    assert isinstance(description, str)
    assert len(description) > 0


# Async Search Functionality Tests
@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_single_query_success(mock_async_ddgs):
    """Test successful execution of a single async search query"""
    mock_results = [
        {
            "title": "Async Example 1",
            "href": "https://example1.com",
            "body": "First async test snippet",
        },
        {
            "title": "Async Example 2",
            "href": "https://example2.com",
            "body": "Second async test snippet",
        },
    ]
    mock_ddgs_instance = AsyncMock()
    mock_ddgs_instance.atext.return_value = mock_results
    mock_async_ddgs.return_value = mock_ddgs_instance

    search = AsyncDuckDuckGoSearch(queries=["test query"])
    result = await search.call()

    assert "Async Example 1" in result
    assert "https://example1.com" in result
    assert "First async test snippet" in result
    mock_ddgs_instance.atext.assert_called_once_with(
        "test query", max_results=search._get_config().max_results_per_query
    )


@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_multiple_queries_success(mock_async_ddgs):
    """Test successful execution of multiple async search queries"""
    mock_results = [
        {
            "title": "Async Multiple Result 1",
            "href": "https://example1.com",
            "body": "First async query result",
        },
        {
            "title": "Async Multiple Result 2",
            "href": "https://example2.com",
            "body": "Second async query result",
        },
    ]
    mock_ddgs_instance = AsyncMock()
    mock_ddgs_instance.atext.return_value = mock_results
    mock_async_ddgs.return_value = mock_ddgs_instance

    search = AsyncDuckDuckGoSearch(queries=["query1", "query2"])
    result = await search.call()

    assert "Async Multiple Result 1" in result
    assert "https://example1.com" in result
    assert mock_ddgs_instance.atext.call_count == 2


@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_empty_results(mock_async_ddgs):
    """Test handling of empty async search results"""
    mock_ddgs_instance = AsyncMock()
    mock_ddgs_instance.atext.return_value = []
    mock_async_ddgs.return_value = mock_ddgs_instance

    search = AsyncDuckDuckGoSearch(queries=["test query"])
    result = await search.call()

    assert result == ""


# Async Error Handling Tests
@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_ddgs_initialization_error(mock_async_ddgs):
    """Test error handling during async DDGS initialization"""
    mock_async_ddgs.side_effect = Exception("AsyncDDGS initialization failed")

    search = AsyncDuckDuckGoSearch(queries=["test query"])
    result = await search.call()

    assert "Failed to search the web for text" in result


@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_search_execution_error(mock_async_ddgs):
    """Test error handling during async search execution"""
    mock_ddgs_instance = AsyncMock()
    mock_ddgs_instance.atext.side_effect = Exception("Search execution failed")
    mock_async_ddgs.return_value = mock_ddgs_instance

    search = AsyncDuckDuckGoSearch(queries=["test query"])
    result = await search.call()

    assert "Failed to search the web for text" in result


@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_url_parsing_error(mock_async_ddgs):
    """Test error handling during async URL parsing"""
    mock_results = [
        {
            "title": "Async Invalid URL",
            "href": None,
            "body": "Test async snippet with invalid URL",
        },
        {
            "title": "Async Valid URL",
            "href": "https://example.com",
            "body": "Test async snippet with valid URL",
        },
    ]
    mock_ddgs_instance = AsyncMock()
    mock_ddgs_instance.atext.return_value = mock_results
    mock_async_ddgs.return_value = mock_ddgs_instance

    search = AsyncDuckDuckGoSearch(queries=["test query"])
    result = await search.call()

    assert (
        "Title: Async Invalid URL\n"
        "URL: None\n"
        "Snippet: Test async snippet with invalid URL\n"
        "\n"
        "Title: Async Valid URL\n"
        "URL: https://example.com\n"
        "Snippet: Test async snippet with valid URL"
    ) in result


# Async Custom Configuration Tests
@mark.asyncio
@patch("mirascope.tools.web._duckduckgo.AsyncDDGS")
async def test_async_custom_max_results(mock_async_ddgs):
    """Test using custom maximum results configuration in async search"""
    custom_config = DuckDuckGoSearchConfig(max_results_per_query=5)
    CustomAsyncSearch = AsyncDuckDuckGoSearch.from_config(custom_config)

    mock_ddgs_instance = AsyncMock()
    mock_ddgs_instance.atext.return_value = []
    mock_async_ddgs.return_value = mock_ddgs_instance

    search = CustomAsyncSearch(queries=["test query"])  # pyright: ignore [reportCallIssue]
    await search.call()

    mock_ddgs_instance.atext.assert_called_once_with("test query", max_results=5)


# Async Integration Tests
@mark.asyncio
async def test_async_integration_with_search_tool_base():
    """Test integration with async search tool base class"""
    search = AsyncDuckDuckGoSearch(queries=["test"])
    description = search.usage_description()

    assert isinstance(description, str)
    assert len(description) > 0



"""Utility for setting up a provider-specific call."""

import inspect
from collections.abc import (
    Awaitable,
    Callable,
    Sequence,
)
from typing import Any, Protocol, TypeVar, cast

from ..call_kwargs import BaseCallKwargs
from ..call_params import BaseCallParams, CommonCallParams
from ..dynamic_config import BaseDynamicConfig
from ..message_param import BaseMessageParam
from ..tool import BaseTool
from . import get_prompt_template, parse_prompt_messages
from ._convert_base_model_to_base_tool import convert_base_model_to_base_tool
from ._convert_function_to_base_tool import convert_function_to_base_tool

_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams, covariant=True)
_CALL_PARAMS_KEYS = set(CommonCallParams.__annotations__)


class ConvertCommonParamsFunc(Protocol[_BaseCallParamsT]):
    def __call__(self, common_params: CommonCallParams) -> _BaseCallParamsT: ...


def setup_call(
    fn: Callable[..., _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]]
    | Callable[..., Sequence[BaseMessageParam]]
    | Callable[..., Awaitable[Sequence[BaseMessageParam]]],
    fn_args: dict[str, Any],
    dynamic_config: _BaseDynamicConfigT,
    tools: Sequence[type[BaseTool] | Callable] | None,
    tool_type: type[_BaseToolT],
    call_params: _BaseCallParamsT | CommonCallParams,
    convert_common_call_params: ConvertCommonParamsFunc[_BaseCallParamsT],
) -> tuple[
    str | None,
    list[BaseMessageParam | Any],
    list[type[_BaseToolT]] | None,
    BaseCallKwargs,
]:
    if isinstance(call_params, dict) and call_params.keys() <= _CALL_PARAMS_KEYS:
        call_params = convert_common_call_params(cast(CommonCallParams, call_params))
    call_kwargs = cast(BaseCallKwargs[_BaseToolT], dict(call_params))
    prompt_template, messages = None, None
    if dynamic_config is not None:
        tools = dynamic_config.get("tools", tools)
        messages = dynamic_config.get("messages", None)
        if messages is not None and not isinstance(messages, list):
            messages = list(messages)
        dynamic_call_params = dynamic_config.get("call_params", None)
        if dynamic_call_params:
            call_kwargs |= dynamic_call_params

    if not messages:
        prompt_template = get_prompt_template(fn)
        assert prompt_template is not None, "The function must have a prompt template."
        messages = parse_prompt_messages(
            roles=["system", "user", "assistant"],
            template=prompt_template,
            attrs=fn_args,
            dynamic_config=dynamic_config,
        )

    tool_types = None
    if tools:
        tool_types = [
            convert_base_model_to_base_tool(tool, tool_type)
            if inspect.isclass(tool)
            else convert_function_to_base_tool(tool, tool_type)
            for tool in tools
        ]
        call_kwargs["tools"] = [tool_type.tool_schema() for tool_type in tool_types]

    return prompt_template, messages, tool_types, call_kwargs



"""Tests the `cohere._utils.convert_message_params` function."""

import pytest
from cohere.types import ChatMessage

from mirascope.core.base import AudioPart, BaseMessageParam, ImagePart, TextPart
from mirascope.core.cohere._utils._convert_message_params import convert_message_params


def test_convert_message_params() -> None:
    """Tests the `convert_message_params` function."""

    message_params: list[BaseMessageParam | ChatMessage] = [
        ChatMessage(message="Hello"),
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Hello"),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params[0].dict() == ChatMessage(message="Hello").dict()
    assert (
        converted_message_params[1].dict()
        == ChatMessage(message="Hello", role="USER").dict()  # type: ignore
    )
    assert (
        converted_message_params[2].dict()
        == ChatMessage(message="Hello", role="USER").dict()  # type: ignore
    )

    with pytest.raises(
        ValueError,
        match="Cohere currently only supports text parts.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail="auto",
                        )
                    ],
                )
            ]
        )

    with pytest.raises(
        ValueError,
        match="Cohere currently only supports text parts.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(type="audio", media_type="audio/mp3", audio=b"audio")
                    ],
                )
            ]
        )



from mirascope.core.base._utils._convert_provider_finish_reason_to_finish_reason import (
    FinishReasonMappingValue,
    _convert_finish_reasons_to_common_finish_reasons_from_mapping,
)
from mirascope.core.base.types import FinishReason

_FinishReasonMapping: dict[str, FinishReasonMappingValue] = {
    "COMPLETE": "stop",
    "STOP_SEQUENCE": "stop",
    "ERROR": "stop",
    "ERROR_TOXIC": "content_filter",
    "ERROR_LIMIT": "length",
    "USER_CANCEL": "stop",
    "MAX_TOKENS": "length",
}


def _convert_finish_reasons_to_common_finish_reasons(
    finish_reasons: list[str] | None,
) -> list[FinishReason] | None:
    """Provider-agnostic finish reasons."""
    return _convert_finish_reasons_to_common_finish_reasons_from_mapping(
        finish_reasons, _FinishReasonMapping
    )



"""Tests the `cohere._utils.setup_call` module."""

import os
from collections.abc import Iterator
from unittest.mock import MagicMock, patch

import pytest
from cohere import NonStreamedChatResponse
from cohere.types import ChatMessage
from pydantic import BaseModel

from mirascope.core.cohere._utils._convert_common_call_params import (
    convert_common_call_params,
)
from mirascope.core.cohere._utils._setup_call import setup_call
from mirascope.core.cohere.tool import CohereTool

os.environ["CO_API_KEY"] = "test"


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@patch(
    "mirascope.core.cohere._utils._setup_call.Client.chat_stream",
    new_callable=MagicMock,
)
@patch("mirascope.core.cohere._utils._setup_call.Client.chat", new_callable=MagicMock)
@patch(
    "mirascope.core.cohere._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.cohere._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_chat: MagicMock,
    mock_chat_stream: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_chat.__name__ = "chat"
    mock_chat_stream.__name__ = "chat_stream"
    mock_utils.setup_call = mock_base_setup_call
    system_message = "system"
    preamble_message = "preamble"
    mock_base_setup_call.return_value[-1] = {
        "preamble": preamble_message,
    }
    mock_convert_message_params.return_value = [
        ChatMessage(role="SYSTEM", message=system_message),  # type: ignore
        ChatMessage(role="USER", message="history user"),  # type: ignore
        ChatMessage(
            role="CHATBOT",  # type: ignore
            message="history system",
        ),
        ChatMessage(
            role="USER",  # type: ignore
            message="content",
        ),
    ]
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="command-r-plus",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "model" in call_kwargs and call_kwargs["model"] == "command-r-plus"
    assert "message" in call_kwargs and call_kwargs["message"] == messages[-1].message
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, CohereTool, {}, convert_common_call_params
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value
    assert "preamble" in call_kwargs
    assert "chat_history" in call_kwargs
    assert call_kwargs["preamble"] == f"{preamble_message}\n\n{system_message}"
    assert call_kwargs["chat_history"] == mock_convert_message_params.return_value[:-1]
    assert (
        call_kwargs["message"] == mock_convert_message_params.return_value[-1].message
    )
    mock_chat.return_value = MagicMock(spec=NonStreamedChatResponse)
    chat = create(stream=False, **call_kwargs)
    stream = create(stream=True, **call_kwargs)
    assert isinstance(chat, NonStreamedChatResponse)
    assert isinstance(stream, Iterator)


@patch(
    "mirascope.core.cohere._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.cohere._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content.return_value = "content"
    mock_base_setup_call.return_value[1] = [
        ChatMessage(role="user", message="")  # type: ignore
    ]
    mock_base_setup_call.return_value[-1]["tools"] = MagicMock()
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="command-r-plus",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[-1] == ChatMessage(
        role="user",  # type: ignore
        message=mock_utils.json_mode_content.return_value,
        tool_calls=None,
    )
    assert "tools" in call_kwargs


@patch(
    "mirascope.core.cohere._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.cohere._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with extraction."""
    mock_utils.setup_call = mock_base_setup_call
    _, _, messages, _, call_kwargs = setup_call(
        model="command-r-plus",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=BaseModel,
        stream=False,
    )
    assert "model" in call_kwargs
    assert "message" in call_kwargs
    assert call_kwargs["model"] == "command-r-plus"
    assert call_kwargs["message"] == messages[-1].message



"""The `AzureTool` class for easy tool usage with Azure LLM calls.

usage docs: learn/tools.md
"""

from __future__ import annotations

import jiter
from azure.ai.inference.models import (
    ChatCompletionsToolCall,
    ChatCompletionsToolDefinition,
    FunctionDefinition,
)
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool, GenerateJsonSchemaNoTitles, ToolConfig


class GenerateAzureStrictToolJsonSchema(GenerateJsonSchemaNoTitles):
    _azure_strict = True


class AzureToolConfig(ToolConfig, total=False):
    """A tool configuration for Azure-specific features."""

    strict: bool


class AzureTool(BaseTool):
    """A class for defining tools for Azure LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.azure import azure_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @azure_call("gpt-4o-mini", tools=[format_book])
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `AzureTool` instance
        print(tool.call())
    ```
    """

    __provider__ = "azure"
    __tool_config_type__ = AzureToolConfig

    tool_call: SkipJsonSchema[ChatCompletionsToolCall]

    @classmethod
    def tool_schema(cls) -> ChatCompletionsToolDefinition:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.azure import AzureTool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = AzureTool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the Azure-specific tool schema
        ```
        """
        fn = FunctionDefinition(name=cls._name(), description=cls._description())
        schema_generator = GenerateJsonSchemaNoTitles
        model_schema = cls.model_json_schema(schema_generator=schema_generator)
        if model_schema["properties"]:
            fn["parameters"] = model_schema
        return ChatCompletionsToolDefinition(function=fn)

    @classmethod
    def from_tool_call(cls, tool_call: ChatCompletionsToolCall) -> AzureTool:
        """Constructs an `AzureTool` instance from a `tool_call`.

        Args:
            tool_call: The Azure tool call from which to construct this tool instance.
        """
        model_json = {"tool_call": tool_call}
        if args := tool_call.function.arguments:
            model_json |= jiter.from_json(args.encode())
        return cls.model_validate(model_json)



"""Mirascope x HyperDX Integration."""

import os
from collections.abc import Callable
from typing import ParamSpec, TypeVar

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
)

from mirascope.integrations.otel._utils import configure
from mirascope.integrations.otel._with_otel import with_otel

_P = ParamSpec("_P")
_R = TypeVar("_R")


def with_hyperdx() -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Decorator to wrap a function with hyperdx.

    Example:

    ```python
    import os

    from mirascope.core import anthropic, prompt_template
    from mirascope.integrations.otel import with_hyperdx

    os.environ["HYPERDX_API_KEY"] = "YOUR_API_KEY"


    def format_book(title: str, author: str):
        return f"{title} by {author}"


    @with_hyperdx()
    @anthropic.call(model="claude-3-5-sonnet-20240620", tools=[format_book])
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    print(recommend_book("fantasy"))
    ```
    """
    provider = trace.get_tracer_provider()
    if not isinstance(provider, TracerProvider):
        configure(
            processors=[
                BatchSpanProcessor(
                    OTLPSpanExporter(
                        endpoint="https://in-otel.hyperdx.io/v1/traces",
                        headers={"authorization": os.getenv("HYPERDX_API_KEY", "")},
                    )
                )
            ]
        )
    return with_otel()
